# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the pydantic wrapping for REST server
"""
# Standard
from typing import Dict, List, Union, get_args
import enum

# Third Party
import numpy as np
import pydantic
import pytest

# First Party
from py_to_proto.dataclass_to_proto import Annotated

# Local
from caikit.core import DataObjectBase, dataobject
from caikit.core.data_model.base import DataBase
from caikit.interfaces.common.data_model.stream_sources import File
from caikit.interfaces.nlp.data_model.text_generation import (
    GeneratedTextStreamResult,
    GeneratedToken,
)
from caikit.runtime.http_server.pydantic_wrapper import (
    PYDANTIC_TO_DM_MAPPING,
    _from_base64,
    _get_pydantic_type,
    dataobject_to_pydantic,
    pydantic_to_dataobject,
)
from caikit.runtime.service_generation.data_stream_source import make_data_stream_source
from sample_lib.data_model.sample import (
    SampleInputType,
    SampleOutputType,
    SampleTrainingType,
)


def test_pydantic_to_dataobject_simple():
    """Test building our simple DM objects through pydantic objects"""
    # get our DM class
    sample_input_dm_class = DataBase.get_class_for_name("SampleInputType")
    # Create pydantic model for our DM class
    sample_input_pydantic_model = dataobject_to_pydantic(sample_input_dm_class)
    # build our DM object using a pydantic object
    sample_input_dm_obj = pydantic_to_dataobject(
        sample_input_pydantic_model(name="Hello world")
    )

    # assert it's our DM object, all fine and dandy
    assert isinstance(sample_input_dm_obj, DataBase)
    assert sample_input_dm_obj.to_json() == '{"name": "Hello world"}'


def test_pydantic_to_dataobject_datastream_jsondata():
    """Test building our datastream DM objects through pydantic objects"""

    make_data_stream_source(SampleTrainingType)
    # get our DM class
    datastream_dm_class = DataBase.get_class_for_name(
        "DataStreamSourceSampleTrainingType"
    )
    # create pydantic model for our DM class
    datastream_pydantic_model = dataobject_to_pydantic(datastream_dm_class)
    # build our DM Datastream JsonData object using a pydantic object
    datastream_dm_obj = pydantic_to_dataobject(
        datastream_pydantic_model(
            data_stream={
                "data": [{"number": 1, "label": "foo"}, {"number": 2, "label": "bar"}]
            }
        )
    )

    # assert it's our DM object, all fine and dandy
    assert isinstance(datastream_dm_obj, DataBase)
    assert (
        datastream_dm_obj.to_json()
        == '{"jsondata": {"data": [{"number": 1, "label": "foo"}, {"number": 2, "label": "bar"}]}}'
    )


def test_pydantic_to_dataobject_datastream_file():
    make_data_stream_source(SampleTrainingType)
    # get our DM class
    datastream_dm_class = DataBase.get_class_for_name(
        "DataStreamSourceSampleTrainingType"
    )
    # create pydantic model for our DM class
    datastream_pydantic_model = dataobject_to_pydantic(datastream_dm_class)

    # build our DM Datastream File object using a pydantic object
    datastream_dm_obj = pydantic_to_dataobject(
        datastream_pydantic_model(data_stream={"filename": "hello"})
    )

    # assert it's our DM object, all fine and dandy
    assert isinstance(datastream_dm_obj, DataBase)
    assert isinstance(datastream_dm_obj.data_stream, File)
    assert datastream_dm_obj.to_json() == '{"file": {"filename": "hello"}}'


@pytest.mark.parametrize(
    "input, output",
    [
        (np.integer, int),
        (np.floating, float),
        (int, int),
        (float, float),
        (bool, bool),
        (str, str),
        (
            bytes,
            Annotated[
                bytes, pydantic.functional_validators.BeforeValidator(_from_base64)
            ],
        ),
        (type(None), type(None)),
        (enum.Enum, enum.Enum),
        (Annotated[str, "blah"], str),
        (Union[str, int], Union[str, int]),
        (List[str], List[str]),
        (List[Annotated[str, "blah"]], List[str]),
        (Dict[str, int], Dict[str, int]),
        (Dict[Annotated[str, "blah"], int], Dict[str, int]),
    ],
)
def test_get_pydantic_type(input, output):
    assert _get_pydantic_type(input) == output


def test_get_pydantic_type_union():
    union_type = Union[SampleInputType, SampleOutputType]
    return_type = _get_pydantic_type(union_type)
    assert all(
        issubclass(ret_type, pydantic.BaseModel) for ret_type in get_args(return_type)
    )


def test_get_pydantic_type_DM():
    # DM case
    sample_input_dm_class = DataBase.get_class_for_name("SampleInputType")
    sample_input_pydantic_model = _get_pydantic_type(sample_input_dm_class)

    assert issubclass(sample_input_pydantic_model, pydantic.BaseModel)
    assert sample_input_pydantic_model in PYDANTIC_TO_DM_MAPPING
    assert sample_input_pydantic_model is PYDANTIC_TO_DM_MAPPING.get(
        sample_input_dm_class
    )


def test_get_pydantic_type_throws_random_type():
    # error case
    with pytest.raises(TypeError):
        _get_pydantic_type("some_random_type")


def test_pydantic_wrapping_with_enums():
    """Check that the pydantic wrapping works on our data models when they have enums"""
    # The NLP GeneratedTextStreamResult data model contains enums

    # Check that our data model is fine and dandy
    token = GeneratedToken(text="foo")
    assert token.text == "foo"

    # Wrap the containing data model in pydantic
    dataobject_to_pydantic(GeneratedTextStreamResult)

    # Check that our data model is _still_ fine and dandy
    token = GeneratedToken(text="foo")
    assert token.text == "foo"


def test_pydantic_wrapping_with_lists():
    """Check that pydantic wrapping works on data models with lists"""

    @dataobject(package="http")
    class BarTest(DataObjectBase):
        baz: int

    @dataobject(package="http")
    class FooTest(DataObjectBase):
        bars: List[BarTest]

    foo = FooTest(bars=[BarTest(1)])
    assert foo.bars[0].baz == 1

    dataobject_to_pydantic(FooTest)

    foo = FooTest(bars=[BarTest(1)])
    assert foo.bars[0].baz == 1


def test_dataobject_to_pydantic_simple_DM():
    """Test that we can create a pydantic model from a simple DM"""
    sample_input_dm_class = DataBase.get_class_for_name("SampleInputType")
    sample_input_pydantic_model = dataobject_to_pydantic(sample_input_dm_class)
    model_instance = sample_input_pydantic_model.model_validate_json(
        '{"name": "world"}'
    )
    assert {"name": str} == sample_input_pydantic_model.__annotations__
    assert issubclass(sample_input_pydantic_model, pydantic.BaseModel)
    assert sample_input_pydantic_model in PYDANTIC_TO_DM_MAPPING
    assert model_instance.name == "world"


def test_dataobject_to_pydantic_simple_DM_extra_forbidden_throws():
    """Test that if we forbid extra values, then we raise if we pass in extra values"""
    sample_input_dm_class = DataBase.get_class_for_name("SampleInputType")

    sample_input_pydantic_model = dataobject_to_pydantic(sample_input_dm_class)
    # sample_input_pydantic_model_extra_forbidden doesn't allow anything extra
    with pytest.raises(pydantic.ValidationError) as e1:
        sample_input_pydantic_model.model_validate_json('{"blah": "world"}')
    assert "Extra inputs are not permitted" in e1.value.errors()[0]["msg"]

    with pytest.raises(pydantic.ValidationError) as e2:
        sample_input_pydantic_model.model_validate_json(
            '{"name": "world", "blah": "world"}'
        )
    assert "Extra inputs are not permitted" in e2.value.errors()[0]["msg"]


def test_dataobject_to_pydantic_oneof():
    """Test that we can create a pydantic model from a DM with a Union"""
    make_data_stream_source(SampleTrainingType)
    sample_input_dm_class = DataBase.get_class_for_name(
        "DataStreamSourceSampleTrainingType"
    )
    data_stream_source_pydantic_model = dataobject_to_pydantic(sample_input_dm_class)

    assert issubclass(data_stream_source_pydantic_model, pydantic.BaseModel)
    assert {
        "data_stream": Union[
            PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.JsonData),
            PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.File),
            PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.ListOfFiles),
            PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.Directory),
            PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.S3Files),
        ]
    } == data_stream_source_pydantic_model.__annotations__
    assert data_stream_source_pydantic_model in PYDANTIC_TO_DM_MAPPING

    assert issubclass(
        PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.JsonData),
        type(
            data_stream_source_pydantic_model.model_validate_json(
                '{"data_stream": {"data": [{"number": 1}]}}'
            ).data_stream
        ),
    )
    assert issubclass(
        PYDANTIC_TO_DM_MAPPING.get(sample_input_dm_class.File),
        type(
            data_stream_source_pydantic_model.model_validate_json(
                '{"data_stream": {"filename": "file1"}}'
            ).data_stream
        ),
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Have pylint ignore Class XXXX has no YYYY member so that we can use gRPC enums.
# pylint: disable=E1101
# Standard
from dataclasses import dataclass
from unittest import mock
import json
import os
import signal
import tempfile
import threading
import time
import uuid

# Third Party
from google.protobuf.descriptor_pool import DescriptorPool
from grpc._utilities import RpcMethodHandler
from grpc_health.v1 import health_pb2, health_pb2_grpc
from grpc_reflection.v1alpha.proto_reflection_descriptor_database import (
    ProtoReflectionDescriptorDatabase,
)
import grpc
import pytest
import tls_test_tools

# First Party
import alog

# Local
from caikit import get_config
from caikit.core import MODEL_MANAGER
from caikit.core.data_model.producer import ProducerId
from caikit.interfaces.runtime.data_model import (
    TrainingInfoRequest,
    TrainingJob,
    TrainingStatus,
    TrainingStatusResponse,
)
from caikit.runtime import get_inference_request, get_train_params, get_train_request
from caikit.runtime.grpc_server import RuntimeGRPCServer
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.protobufs import (
    model_runtime_pb2,
    model_runtime_pb2_grpc,
    process_pb2,
    process_pb2_grpc,
)
from caikit.runtime.service_factory import ServicePackage, ServicePackageFactory
from caikit.runtime.utils.servicer_util import build_caikit_library_request_dict
from sample_lib import (
    CompositeModule,
    InnerModule,
    OtherModule,
    SampleModule,
    SamplePrimitiveModule,
    StreamingModule,
)
from sample_lib.data_model import (
    OtherOutputType,
    SampleInputType,
    SampleOutputType,
    SampleTrainingType,
)
from tests.conftest import random_test_id, temp_config
from tests.core.helpers import *
from tests.fixtures import Fixtures
from tests.runtime.conftest import (
    ModuleSubproc,
    register_trained_model,
    runtime_grpc_test_server,
)
import caikit.interfaces.common

## Helpers #####################################################################

log = alog.use_channel("TEST-SERVE-I")

HAPPY_PATH_INPUT_DM = SampleInputType(name="Gabe")
HAPPY_PATH_INPUT = HAPPY_PATH_INPUT_DM.to_proto()
HAPPY_PATH_RESPONSE = SampleOutputType(greeting="Hello Gabe").to_proto()
HAPPY_PATH_TRAIN_RESPONSE = TrainingJob(
    model_name="dummy name", training_id="dummy id"
).to_proto()


def assert_training_successful(
    actual_response, expected, model_name, training_management_stub
):
    assert dir(actual_response) == dir(expected)
    assert actual_response.training_id is not None
    assert isinstance(actual_response.training_id, str)
    assert actual_response.model_name == model_name

    MODEL_MANAGER.get_model_future(actual_response.training_id).wait()

    training_info_request = TrainingInfoRequest(training_id=actual_response.training_id)
    training_management_response: TrainingStatusResponse = (
        TrainingStatusResponse.from_proto(
            training_management_stub.GetTrainingStatus(training_info_request.to_proto())
        )
    )
    status = training_management_response.state
    assert status == TrainingStatus.COMPLETED.value


## Tests #######################################################################


def test_model_train(runtime_grpc_server):
    """Test model train's RUN function"""
    model_train_stub = process_pb2_grpc.ProcessStub(
        runtime_grpc_server.make_local_channel()
    )
    training_id = str(uuid.uuid4())
    model_name = "abc"
    model_train_request = process_pb2.ProcessRequest(
        trainingID=training_id,
        request_dict={
            "train_module": "00110203-0405-0607-0809-0a0b02dd0e0f",
            "training_params": json.dumps(
                {
                    "model_name": model_name,
                    "parameters": {
                        "training_data": {
                            "jsondata": {
                                "data": [SampleTrainingType(number=1).to_dict()]
                            },
                        },
                    },
                }
            ),
        },
        training_input_dir="training_input_dir",
        training_output_dir=os.path.join(
            runtime_grpc_server.workdir, "training_output", training_id
        ),
    )
    training_response = model_train_stub.Run(model_train_request)
    assert isinstance(training_response, process_pb2.ProcessResponse)
    assert training_response == process_pb2.ProcessResponse(
        trainingID=model_train_request.trainingID,
        customTrainingID=model_train_request.customTrainingID,
    )

    # check result of training

    # construct training request for training manager
    training_info_request = TrainingInfoRequest(training_id=training_id).to_proto()

    training_management_service: ServicePackage = (
        ServicePackageFactory().get_service_package(
            ServicePackageFactory.ServiceType.TRAINING_MANAGEMENT,
        )
    )

    training_management_stub = training_management_service.stub_class(
        runtime_grpc_server.make_local_channel()
    )

    # MT training is sync, so it should be COMPLETE immediately
    response: TrainingStatusResponse = TrainingStatusResponse.from_proto(
        training_management_stub.GetTrainingStatus(training_info_request)
    )
    assert response.state == TrainingStatus.COMPLETED.value

    # Make sure we wait for training to finish
    model_future = MODEL_MANAGER.get_model_future(response.training_id)
    result = model_future.load()

    # Make sure we put the path bits in the right order: base/training_id/model_name
    training_id_and_model_name_path_bit = os.path.join(training_id, model_name)
    assert training_id_and_model_name_path_bit in model_future.save_path

    assert (
        result.MODULE_CLASS
        == "sample_lib.modules.sample_task.sample_implementation.SampleModule"
    )
    # Fields with defaults have expected values
    assert result.batch_size == 64
    assert result.learning_rate == 0.0015


def test_components_preinitialized(reset_globals, open_port):
    """Test that all model management components get pre-initialized when the
    server is instantiated
    """
    assert not MODEL_MANAGER._trainers
    assert not MODEL_MANAGER._finders
    assert not MODEL_MANAGER._initializers
    with runtime_grpc_test_server(
        open_port,
    ):
        assert MODEL_MANAGER._trainers
        assert MODEL_MANAGER._finders
        assert MODEL_MANAGER._initializers


def test_predict_sample_module_ok_response(
    sample_task_model_id, runtime_grpc_server, sample_inference_service
):
    """Test RPC CaikitRuntime.SampleTaskPredict successful response"""
    stub = sample_inference_service.stub_class(runtime_grpc_server.make_local_channel())
    predict_request = get_inference_request(SampleModule.TASK_CLASS)(
        sample_input=HAPPY_PATH_INPUT_DM
    ).to_proto()

    actual_response = stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", sample_task_model_id)]
    )

    assert actual_response == HAPPY_PATH_RESPONSE


def test_global_predict_build_caikit_library_request_dict_creates_caikit_core_run_kwargs(
    sample_inference_service,
):
    """Test using proto versus pythonic data model for inference requests to compare diffs"""
    # protobuf request
    proto_request = sample_inference_service.messages.SampleTaskRequest(
        sample_input=HAPPY_PATH_INPUT_DM.to_proto(),
    )
    proto_request_dict = build_caikit_library_request_dict(
        proto_request,
        SampleModule.RUN_SIGNATURE,
    )

    # unset fields not included
    proto_expected_arguments = {"sample_input"}
    assert proto_request.HasField("throw") is False
    assert proto_expected_arguments == set(proto_request_dict.keys())

    # pythonic data model request
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    python_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()

    python_sample_module_request_dict = build_caikit_library_request_dict(
        python_request,
        SampleModule.RUN_SIGNATURE,
    )

    # unset fields are included if they have defaults set
    python_expected_arguments = {"sample_input", "throw"}
    assert python_request.HasField("throw") is True
    assert python_expected_arguments == set(python_sample_module_request_dict.keys())


def test_predict_streaming_module(
    streaming_task_model_id, runtime_grpc_server, sample_inference_service
):
    """Test RPC CaikitRuntime.StreamingTaskPredict successful response"""
    stub = sample_inference_service.stub_class(runtime_grpc_server.make_local_channel())
    predict_class = get_inference_request(
        StreamingModule.TASK_CLASS, input_streaming=False, output_streaming=True
    )
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()

    stream = stub.ServerStreamingStreamingTaskPredict(
        predict_request, metadata=[("mm-model-id", streaming_task_model_id)]
    )

    count = 0
    for response in stream:
        assert response == HAPPY_PATH_RESPONSE
        count += 1
    assert count == 10


def test_predict_sample_module_error_response(
    runtime_grpc_server, sample_inference_service
):
    """Test RPC CaikitRuntime.SampleTaskPredict error response"""
    with pytest.raises(grpc.RpcError) as context:
        stub = sample_inference_service.stub_class(
            runtime_grpc_server.make_local_channel()
        )
        predict_class = get_inference_request(SampleModule.TASK_CLASS)
        predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()

        stub.SampleTaskPredict(
            predict_request, metadata=[("mm-model-id", "random_model_id")]
        )
    assert context.value.code() == grpc.StatusCode.NOT_FOUND


@pytest.mark.skip("Skipping for now since we're doing streaming stuff")
def test_rpc_validation_on_predict(
    sample_task_model_id, runtime_grpc_server, sample_inference_service
):
    """Check that the server catches models sent to the wrong task RPCs"""
    stub = sample_inference_service.stub_class(runtime_grpc_server.make_local_channel())
    predict_class = get_inference_request(OtherModule.TASK_CLASS)
    predict_request = predict_class(
        sample_input_sampleinputtype=HAPPY_PATH_INPUT_DM
    ).to_proto()

    with pytest.raises(
        grpc.RpcError,
        match="Wrong inference RPC invoked for model class .* Use SampleTaskPredict instead of OtherTaskPredict",
    ) as context:
        stub.OtherTaskPredict(
            predict_request, metadata=[("mm-model-id", sample_task_model_id)]
        )
    assert context.value.code() == grpc.StatusCode.INVALID_ARGUMENT


def test_rpc_validation_on_predict_for_unsupported_model(
    runtime_grpc_server: RuntimeGRPCServer, sample_inference_service, tmp_path
):
    """Check that the server catches models that have no supported inference rpc"""
    unsupported_model = InnerModule()
    tmpdir = str(tmp_path)
    unsupported_model.save(tmpdir)
    model_id = random_test_id()
    try:
        runtime_grpc_server._global_predict_servicer._model_manager.load_model(
            model_id, tmpdir, "foo"
        )

        stub = sample_inference_service.stub_class(
            runtime_grpc_server.make_local_channel()
        )
        predict_class = get_inference_request(SampleModule.TASK_CLASS)
        predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
        with pytest.raises(grpc.RpcError) as context:
            stub.SampleTaskPredict(
                predict_request, metadata=[("mm-model-id", model_id)]
            )
        assert context.value.code() == grpc.StatusCode.INVALID_ARGUMENT
        assert "Inference for model class" in str(context.value)
        assert "not supported by this runtime" in str(context.value)

    finally:
        runtime_grpc_server._global_predict_servicer._model_manager.unload_model(
            model_id
        )


def test_rpc_validation_on_predict_for_wrong_streaming_flavor(
    runtime_grpc_server: RuntimeGRPCServer, sample_inference_service, tmp_path
):
    """Check that the server catches models that have no supported inference rpc"""
    unary_only_model = SamplePrimitiveModule()
    tmpdir = str(tmp_path)
    unary_only_model.save(tmpdir)
    model_id = random_test_id()
    try:
        runtime_grpc_server._global_predict_servicer._model_manager.load_model(
            model_id, tmpdir, "foo"
        )

        stub = sample_inference_service.stub_class(
            runtime_grpc_server.make_local_channel()
        )

        predict_class = get_inference_request(
            SampleModule.TASK_CLASS,
        )
        predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
        with pytest.raises(grpc.RpcError) as context:
            response = stub.ServerStreamingSampleTaskPredict(
                predict_request, metadata=[("mm-model-id", model_id)]
            )
            for r in response:
                # try to read off the stream
                pass

        assert context.value.code() == grpc.StatusCode.INVALID_ARGUMENT
        assert "Model class" in str(context.value)
        assert "does not support ServerStreamingSampleTaskPredict" in str(context.value)

    finally:
        runtime_grpc_server._global_predict_servicer._model_manager.unload_model(
            model_id
        )


####### End-to-end tests for train a model and then predict with it
def test_train_fake_module_ok_response_and_can_predict_with_trained_model(
    train_stub,
    inference_stub,
    runtime_grpc_server,
    sample_train_service,
    sample_inference_service,
    training_management_stub,
):
    """Test RPC CaikitRuntime.SampleTaskSampleModuleTrain successful response"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(
            data=[SampleTrainingType(1), SampleTrainingType(2)]
        )
    )
    model_name = random_test_id()
    train_request = get_train_request(SampleModule)(
        model_name=model_name,
        parameters=get_train_params(SampleModule)(
            training_data=training_data,
            union_list=["str", "sequence"],
        ),
    ).to_proto()

    actual_response = train_stub.SampleTaskSampleModuleTrain(train_request)

    assert_training_successful(
        actual_response, HAPPY_PATH_TRAIN_RESPONSE, model_name, training_management_stub
    )
    register_trained_model(
        runtime_grpc_server, actual_response.model_name, actual_response.training_id
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
    inference_response = inference_stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", actual_response.model_name)]
    )
    assert inference_response == HAPPY_PATH_RESPONSE


def test_train_fake_module_ok_response_with_loaded_model_can_predict_with_trained_model(
    sample_task_model_id,
    runtime_grpc_server,
    train_stub,
    inference_stub,
    sample_train_service,
    sample_inference_service,
    training_management_stub,
):
    """Test RPC CaikitRuntime.WorkflowsSampleTaskSampleWorkflowTrain successful response with a loaded model"""
    sample_model = caikit.interfaces.runtime.data_model.ModelPointer(
        model_id=sample_task_model_id
    )
    model_name = random_test_id()
    train_request = get_train_request(CompositeModule)(
        model_name=model_name,
        parameters=get_train_params(CompositeModule)(sample_block=sample_model),
    ).to_proto()

    actual_response = train_stub.SampleTaskCompositeModuleTrain(train_request)
    assert_training_successful(
        actual_response, HAPPY_PATH_TRAIN_RESPONSE, model_name, training_management_stub
    )
    register_trained_model(
        runtime_grpc_server, actual_response.model_name, actual_response.training_id
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(CompositeModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
    inference_response = inference_stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", actual_response.model_name)]
    )
    assert inference_response == HAPPY_PATH_RESPONSE


def test_train_fake_module_does_not_change_another_instance_model_of_block(
    other_task_model_id,
    runtime_grpc_server,
    sample_int_file,
    train_stub,
    inference_stub,
    training_management_stub,
    sample_train_service,
    sample_inference_service,
):
    """This test: original "stock" OtherModule model has batch size 42
    (See fixtures/models/bar/config.yml).
    We then train a custom OtherModule model with batch size 100,
    then we make a predict to each, they should have the correct batch size"""

    # Train an OtherModule with batch size 100
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    training_data = stream_type(file=stream_type.File(filename=sample_int_file))

    train_request = get_train_request(OtherModule)(
        model_name="Bar Training",
        parameters=get_train_params(OtherModule)(
            sample_input_sampleinputtype=HAPPY_PATH_INPUT_DM,
            batch_size=100,
            training_data=training_data,
        ),
    ).to_proto()
    actual_response = train_stub.OtherTaskOtherModuleTrain(train_request)
    assert_training_successful(
        actual_response,
        HAPPY_PATH_TRAIN_RESPONSE,
        "Bar Training",
        training_management_stub,
    )
    register_trained_model(
        runtime_grpc_server, actual_response.model_name, actual_response.training_id
    )

    # make sure the trained model can run inference, and the batch size 100 was used
    predict_class = get_inference_request(OtherModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
    trained_inference_response = inference_stub.OtherTaskPredict(
        predict_request, metadata=[("mm-model-id", actual_response.model_name)]
    )
    expected_trained_inference_response = OtherOutputType(
        farewell="goodbye: Gabe 100 times", producer_id=ProducerId("other_mod", "1.1.1")
    ).to_proto()
    assert trained_inference_response == expected_trained_inference_response

    # make sure the previously loaded OtherModule model still has batch size 42
    original_inference_response = inference_stub.OtherTaskPredict(
        predict_request, metadata=[("mm-model-id", other_task_model_id)]
    )
    expected_original_inference_response = OtherOutputType(
        farewell="goodbye: Gabe 42 times", producer_id=ProducerId("other_mod", "1.1.1")
    ).to_proto()
    assert original_inference_response == expected_original_inference_response


def test_train_primitive_model(
    runtime_grpc_server,
    train_stub,
    inference_stub,
    training_management_stub,
    sample_train_service,
    sample_inference_service,
):
    """Test that we can make a successful training and inference call to the primitive module using primitive inputs"""

    model_name = "primitive_trained_model"

    train_request = get_train_request(SamplePrimitiveModule)(
        model_name=model_name,
        parameters=get_train_params(SamplePrimitiveModule)(
            sample_input=HAPPY_PATH_INPUT_DM,
            simple_list=["hello", "world"],
            union_list=["str", "sequence"],
            union_list2=[1, 2],
            union_list3=[True, False],
            union_list4=123,
            training_params_json_dict={"foo": {"bar": [1, 2, 3]}},
            training_params_json_dict_list=[{"foo": {"bar": [1, 2, 3]}}],
            training_params_dict={"layer_sizes": 100, "window_scaling": 200},
            training_params_dict_int={1: 0.1, 2: 0.01},
        ),
    ).to_proto()

    training_response = train_stub.SampleTaskSamplePrimitiveModuleTrain(train_request)
    assert_training_successful(
        training_response,
        HAPPY_PATH_TRAIN_RESPONSE,
        model_name,
        training_management_stub,
    )
    register_trained_model(
        runtime_grpc_server, training_response.model_name, training_response.training_id
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()

    inference_response = inference_stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", training_response.model_name)]
    )
    expected_inference_response = SampleOutputType(
        greeting="hello: primitives! [1, 2, 3] 100"
    ).to_proto()
    assert inference_response == expected_inference_response


##### Test different datastream types #####
def test_train_fake_module_ok_response_with_datastream_jsondata(
    runtime_grpc_server,
    train_stub,
    inference_stub,
    sample_train_service,
    sample_inference_service,
    training_management_stub,
):
    """Test RPC CaikitRuntime.SampleTaskSampleModuleTrainRequest successful response with training data json type"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(
            data=[SampleTrainingType(1), SampleTrainingType(2)]
        )
    )
    model_name = random_test_id()

    train_request = get_train_request(SampleModule)(
        model_name=model_name,
        parameters=get_train_params(SampleModule)(
            batch_size=42,
            training_data=training_data,
        ),
    ).to_proto()

    actual_response = train_stub.SampleTaskSampleModuleTrain(train_request)
    assert_training_successful(
        actual_response, HAPPY_PATH_TRAIN_RESPONSE, model_name, training_management_stub
    )
    register_trained_model(
        runtime_grpc_server, actual_response.model_name, actual_response.training_id
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
    inference_response = inference_stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", actual_response.model_name)]
    )
    assert inference_response == HAPPY_PATH_RESPONSE


def test_train_fake_module_ok_response_with_datastream_csv_file(
    runtime_grpc_server,
    train_stub,
    inference_stub,
    sample_train_service,
    sample_inference_service,
    sample_csv_file,
    training_management_stub,
):
    """Test RPC CaikitRuntime.SampleTaskSampleModuleTrainRequest successful response with training data file type"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(file=stream_type.File(filename=sample_csv_file))
    model_name = random_test_id()

    train_request = get_train_request(SampleModule)(
        model_name=model_name,
        parameters=get_train_params(SampleModule)(
            training_data=training_data,
        ),
    ).to_proto()

    actual_response = train_stub.SampleTaskSampleModuleTrain(train_request)
    assert_training_successful(
        actual_response, HAPPY_PATH_TRAIN_RESPONSE, model_name, training_management_stub
    )
    register_trained_model(
        runtime_grpc_server, actual_response.model_name, actual_response.training_id
    )

    # make sure the trained model can run inference
    predict_class = get_inference_request(SampleModule.TASK_CLASS)
    predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
    inference_response = inference_stub.SampleTaskPredict(
        predict_request, metadata=[("mm-model-id", actual_response.model_name)]
    )
    assert inference_response == HAPPY_PATH_RESPONSE


def test_train_and_successfully_cancel_training(
    train_stub, sample_train_service, training_management_stub
):
    # train a model, make sure training doesn't have error
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(
            data=[SampleTrainingType(1), SampleTrainingType(2)]
        )
    )
    model_name = random_test_id()
    # start a training that sleeps for a long time, so I can cancel

    train_request = get_train_request(SampleModule)(
        model_name=model_name,
        parameters=get_train_params(SampleModule)(
            training_data=training_data, sleep_time=10
        ),
    ).to_proto()
    train_response = train_stub.SampleTaskSampleModuleTrain(train_request)

    training_id = train_response.training_id
    training_info_request = TrainingInfoRequest(training_id=training_id)
    training_management_response: TrainingStatusResponse = (
        TrainingStatusResponse.from_proto(
            training_management_stub.GetTrainingStatus(training_info_request.to_proto())
        )
    )
    assert (
        training_management_response.state == TrainingStatus.RUNNING.value
    ), "Could not cancel this training within 10s"
    # cancel the training
    canceled_response = training_management_stub.CancelTraining(
        training_info_request.to_proto()
    )
    assert canceled_response.state == TrainingStatus.CANCELED.value


def test_cancel_does_not_affect_other_models(
    train_stub, sample_train_service, training_management_stub
):
    # train a model, make sure training is running
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        jsondata=stream_type.JsonData(
            data=[SampleTrainingType(1), SampleTrainingType(2)]
        )
    )
    model_name = random_test_id()
    # start a training that sleeps for a long time, so I can cancel
    train_request = get_train_request(SampleModule)(
        model_name=model_name,
        parameters=get_train_params(SampleModule)(
            training_data=training_data, sleep_time=10
        ),
    ).to_proto()
    train_response = train_stub.SampleTaskSampleModuleTrain(train_request)

    assert dir(train_response) == dir(HAPPY_PATH_TRAIN_RESPONSE)
    assert train_response.training_id is not None
    assert isinstance(train_response.training_id, str)
    assert train_response.model_name == model_name

    training_id = train_response.training_id
    training_info_request = TrainingInfoRequest(training_id=training_id)
    training_management_response: TrainingStatusResponse = (
        TrainingStatusResponse.from_proto(
            training_management_stub.GetTrainingStatus(training_info_request.to_proto())
        )
    )

    # train another model
    model_name2 = random_test_id()
    train_request2 = get_train_request(SampleModule)(
        model_name=model_name2,
        parameters=get_train_params(SampleModule)(training_data=training_data),
    ).to_proto()
    train_response2 = train_stub.SampleTaskSampleModuleTrain(train_request2)

    # cancel the first training
    assert (
        training_management_response.state == TrainingStatus.RUNNING.value
    ), "Could not cancel this training within 10s"
    canceled_response = training_management_stub.CancelTraining(
        training_info_request.to_proto()
    )
    assert canceled_response.state == TrainingStatus.CANCELED.value

    # second training should be COMPLETED
    assert_training_successful(
        train_response2,
        HAPPY_PATH_TRAIN_RESPONSE,
        model_name2,
        training_management_stub,
    )


#### Error cases for train tests #####
def test_train_fake_module_error_response_with_unloaded_model(
    train_stub, sample_train_service
):
    """Test RPC CaikitRuntime.SampleTaskCompositeModuleTrain error response because sample model is not loaded"""
    with pytest.raises(grpc.RpcError) as context:
        sample_model = caikit.interfaces.runtime.data_model.ModelPointer(
            model_id=random_test_id()
        )
        train_request = get_train_request(CompositeModule)(
            model_name=random_test_id(),
            parameters=get_train_params(CompositeModule)(sample_block=sample_model),
        ).to_proto()

        train_stub.SampleTaskCompositeModuleTrain(train_request)
    assert context.value.code() == grpc.StatusCode.NOT_FOUND


#### ModelRuntime tests ####
def test_load_model_ok_response(runtime_grpc_server):
    """Test load model's successful response"""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )
    load_model_request = model_runtime_pb2.LoadModelRequest(
        modelId="real",
        modelType="foo",
        modelPath=Fixtures.get_good_model_path(),
        modelKey="bar",
    )
    actual_response = stub.loadModel(load_model_request)
    assert actual_response.sizeInBytes > 0


def test_load_model_notfound_error_response(runtime_grpc_server):
    """Test load model's error response - model does not exist"""
    with pytest.raises(grpc.RpcError) as context:
        stub = model_runtime_pb2_grpc.ModelRuntimeStub(
            runtime_grpc_server.make_local_channel()
        )
        load_model_request = model_runtime_pb2.LoadModelRequest(
            modelId="foo",
            modelType="bar",
            modelPath="test/some/invalid/path",
            modelKey="baz",
        )
        stub.loadModel(load_model_request)
    assert context.value.code() == grpc.StatusCode.NOT_FOUND


def test_load_model_badmodel_error_response(runtime_grpc_server):
    """Test load model's error response - invalid model"""
    with pytest.raises(grpc.RpcError) as context:
        stub = model_runtime_pb2_grpc.ModelRuntimeStub(
            runtime_grpc_server.make_local_channel()
        )
        load_model_request = model_runtime_pb2.LoadModelRequest(
            modelId="foo",
            modelType="bar",
            modelPath=Fixtures.get_bad_model_archive_path(),
            modelKey="baz",
        )
        stub.loadModel(load_model_request)
    assert context.value.code() == grpc.StatusCode.INTERNAL


def test_unload_model_ok_response(sample_task_model_id, runtime_grpc_server):
    """Test unload model's successful response"""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )
    unload_model_request = model_runtime_pb2.UnloadModelRequest(
        modelId=sample_task_model_id
    )
    # Unload model throws on failure (response message has no fields)
    stub.unloadModel(unload_model_request)


def test_unload_model_does_not_throw_if_model_does_not_exist(runtime_grpc_server):
    """Unloading a model that does not exist (or has already been deleted) is totally cool"""
    try:
        stub = model_runtime_pb2_grpc.ModelRuntimeStub(
            runtime_grpc_server.make_local_channel()
        )
        unload_model_request = model_runtime_pb2.UnloadModelRequest(modelId="new_model")
        stub.unloadModel(unload_model_request)
    except grpc.RpcError:
        assert False, "Unload for a model that does not exist threw an error!"


def test_predict_model_size_ok_response(runtime_grpc_server):
    """Test predict model size successful response on model that is not loaded before"""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )
    predict_model_size_request = model_runtime_pb2.PredictModelSizeRequest(
        modelId="foo", modelType="bar", modelPath=Fixtures.get_good_model_path()
    )
    actual_response = stub.predictModelSize(predict_model_size_request)
    assert 0 < actual_response.sizeInBytes


def test_predict_model_size_on_loaded_model_ok_response(
    sample_task_model_id, runtime_grpc_server
):
    """Test predict model size successful response on a model that has been loaded"""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )
    predict_model_size_request = model_runtime_pb2.PredictModelSizeRequest(
        modelId=sample_task_model_id, modelPath=Fixtures.get_good_model_path()
    )
    actual_response = stub.predictModelSize(predict_model_size_request)
    assert 0 < actual_response.sizeInBytes


def test_predict_model_size_model_notfound_error_response(runtime_grpc_server):
    """Test predict model size with unknown model path error response"""
    with pytest.raises(grpc.RpcError) as context:
        stub = model_runtime_pb2_grpc.ModelRuntimeStub(
            runtime_grpc_server.make_local_channel()
        )
        predict_model_size_request = model_runtime_pb2.PredictModelSizeRequest(
            modelId=str(uuid.uuid4()), modelPath="/test/Does/Not/Exist"
        )
        stub.predictModelSize(predict_model_size_request)
    assert context.value.code() == grpc.StatusCode.NOT_FOUND


def test_model_size_ok_response(sample_task_model_id, runtime_grpc_server):
    """Test model size successful response"""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )

    model_size_request = model_runtime_pb2.ModelSizeRequest(
        modelId=sample_task_model_id
    )
    actual_response = stub.modelSize(model_size_request)
    # Mar. 14, 23
    # The size of the directory pointed to by Fixtures.get_good_model_path() is 355 now.
    expected_size = (
        355
        * get_config().inference_plugin.model_mesh.model_size_multipliers[
            Fixtures.get_good_model_type()
        ]
    )
    assert abs(actual_response.sizeInBytes - expected_size) < 100


def test_model_size_model_notfound_error_response(runtime_grpc_server):
    """Test model size with model not found error response"""
    with pytest.raises(grpc.RpcError) as context:
        stub = model_runtime_pb2_grpc.ModelRuntimeStub(
            runtime_grpc_server.make_local_channel()
        )
        model_size_request = model_runtime_pb2.ModelSizeRequest(
            modelId="unknown_model_size_test"
        )
        stub.modelSize(model_size_request)
    assert context.value.code() == grpc.StatusCode.NOT_FOUND


def test_runtime_status_ok_response(runtime_grpc_server):
    """Test run-time status successful response."""
    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )
    runtime_status_request = model_runtime_pb2.RuntimeStatusRequest()
    actual_response = stub.runtimeStatus(runtime_status_request)
    assert actual_response.status == model_runtime_pb2.RuntimeStatusResponse.READY
    assert (
        actual_response.capacityInBytes
        == get_config().inference_plugin.model_mesh.capacity
    )
    assert actual_response.maxLoadingConcurrency == 2
    assert (
        actual_response.modelLoadingTimeoutMs
        == get_config().inference_plugin.model_mesh.model_loading_timeout_ms
    )
    assert actual_response.defaultModelSizeInBytes == 18874368
    assert actual_response.numericRuntimeVersion == 0


#### Health Probe tests ####
def test_grpc_health_probe_ok_response(runtime_grpc_server):
    """Test health check successful response"""
    stub = health_pb2_grpc.HealthStub(runtime_grpc_server.make_local_channel())
    health_check_request = health_pb2.HealthCheckRequest()
    actual_response = stub.Check(health_check_request)
    assert actual_response.status == 1


def test_grpc_server_can_render_all_necessary_protobufs(
    runtime_grpc_server, sample_inference_service, sample_train_service, tmp_path
):
    """Test service protobufs can be rendered"""
    tempdir = str(tmp_path)
    runtime_grpc_server.render_protos(tempdir)
    rendered_protos = os.listdir(tempdir)

    assert (
        f"{sample_inference_service.service.__name__.lower()}.proto" in rendered_protos
    )
    assert f"{sample_train_service.service.__name__.lower()}.proto" in rendered_protos


# Test related to handling load aborts for sad situations :(
def test_canceling_model_loads_causes_exceptions(runtime_grpc_server):
    """Test to make sure that if we cancel a load, an exception is thrown inside our loader."""

    request_received = threading.Event()
    request_finished = threading.Event()

    stub = model_runtime_pb2_grpc.ModelRuntimeStub(
        runtime_grpc_server.make_local_channel()
    )

    load_model_request = model_runtime_pb2.LoadModelRequest(
        modelId="model_load_abort_test",
        modelType="foo",
        modelPath=Fixtures.get_good_model_path(),
        modelKey="bar",
    )

    def never_return(*args, **kwargs):
        request_received.set()
        try:
            while True:
                time.sleep(0.01)
        except Exception as e:
            request_finished.set()
            raise e

    manager_instance = ModelManager.get_instance()
    with mock.patch.object(manager_instance, "load_model", never_return):
        # Call the stub - note that .future makes this thing return a future object that
        # we can wait for (or cancel) that will eventually have a meaningful result.
        load_model_future = stub.loadModel.future(load_model_request)
        request_received.wait()
        # Request has been received; now cancel it
        load_model_future.cancel()

        # Wait for an exception to be raised in our mock, and assert it was
        request_finished.wait(10)
        assert request_finished.is_set()


def test_tls(open_port):
    """Boot up a server with TLS enabled and ping it on a secure channel"""
    ca_key = tls_test_tools.generate_key()[0]
    ca_cert = tls_test_tools.generate_ca_cert(ca_key)
    tls_key, tls_cert = tls_test_tools.generate_derived_key_cert_pair(ca_key)

    tls_config = TLSConfig(
        server=KeyPair(cert=tls_cert, key=tls_key), client=KeyPair(cert="", key="")
    )
    with runtime_grpc_test_server(
        open_port,
        tls_config_override=tls_config,
    ) as server:
        _assert_connection(_make_secure_channel(server, ca_cert))


def test_mtls(open_port):
    """Boot up a server with mTLS enabled and ping it on a secure channel"""
    ca_key = tls_test_tools.generate_key()[0]
    ca_cert = tls_test_tools.generate_ca_cert(ca_key)
    tls_key, tls_cert = tls_test_tools.generate_derived_key_cert_pair(ca_key)

    tls_config = TLSConfig(
        server=KeyPair(cert=tls_cert, key=tls_key), client=KeyPair(cert=ca_cert, key="")
    )
    with runtime_grpc_test_server(
        open_port,
        tls_config_override=tls_config,
    ) as server:
        client_key, client_cert = tls_test_tools.generate_derived_key_cert_pair(ca_key)
        _assert_connection(
            _make_secure_channel(server, ca_cert, client_key, client_cert)
        )

        # Once we know we can connect with mTLS, assert we cannot with plain TLS
        bad_channel = _make_secure_channel(server, ca_cert)
        with pytest.raises(grpc.RpcError):
            stub = health_pb2_grpc.HealthStub(bad_channel)
            health_check_request = health_pb2.HealthCheckRequest()
            stub.Check(health_check_request)


@pytest.mark.parametrize(
    "enabled_services",
    [(True, False), (False, True), (False, False)],
)
def test_services_disabled(open_port, enabled_services):
    """Boot up a server with different combinations of services disabled"""
    enable_inference, enable_training = enabled_services
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "enable_inference": enable_inference,
                    "enable_training": enable_training,
                }
            },
        },
        "merge",
    ):
        with runtime_grpc_test_server(open_port) as server:
            _assert_connection(server.make_local_channel())
            assert server.enable_inference == enable_inference
            assert (server._global_predict_servicer and enable_inference) or (
                server._global_predict_servicer is None and not enable_inference
            )
            assert server.enable_training == enable_training
            assert (server.training_service and enable_training) or (
                server.training_service is None and not enable_training
            )


def test_certs_can_be_loaded_as_files(tmp_path, open_port):
    """mTLS test with all tls configs loaded from files"""
    ca_key = tls_test_tools.generate_key()[0]
    ca_cert = tls_test_tools.generate_ca_cert(ca_key)
    tls_key, tls_cert = tls_test_tools.generate_derived_key_cert_pair(ca_key)

    tmpdir = str(tmp_path)

    tls_cert_path = os.path.join(tmpdir, "tls.crt")
    tls_key_path = os.path.join(tmpdir, "tls.key")
    ca_cert_path = os.path.join(tmpdir, "ca.crt")
    with open(tls_cert_path, "w") as f:
        f.write(tls_cert)
    with open(tls_key_path, "w") as f:
        f.write(tls_key)
    with open(ca_cert_path, "w") as f:
        f.write(ca_cert)

    tls_config = TLSConfig(
        server=KeyPair(cert=tls_cert_path, key=tls_key_path),
        client=KeyPair(cert=ca_cert_path, key=""),
    )
    with runtime_grpc_test_server(
        open_port,
        tls_config_override=tls_config,
    ) as server:
        client_key, client_cert = tls_test_tools.generate_derived_key_cert_pair(ca_key)
        _assert_connection(
            _make_secure_channel(server, ca_cert, client_key, client_cert)
        )


def test_metrics_stored_after_server_interrupt(
    sample_task_model_id, sample_inference_service, open_port
):
    """This tests the gRPC server's behaviour when interrupted"""

    with temp_config({"runtime": {"metering": {"enabled": True}}}, "merge"):
        with runtime_grpc_test_server(open_port) as server:
            stub = sample_inference_service.stub_class(server.make_local_channel())
            predict_class = get_inference_request(SampleModule.TASK_CLASS)
            predict_request = predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto()
            _ = stub.SampleTaskPredict(
                predict_request, metadata=[("mm-model-id", sample_task_model_id)]
            )

            # Interrupt server
            server.interrupt(None, None)

            # Assertions on the created metrics file
            with open(server._global_predict_servicer.rpc_meter.file_path) as f:
                data = [json.loads(line) for line in f]

                assert len(data) == 1
                assert list(data[0].keys()) == [
                    "timestamp",
                    "batch_size",
                    "model_type_counters",
                    "container_id",
                ]
                assert data[0]["batch_size"] == 1
                assert len(data[0]["model_type_counters"]) == 1
                assert data[0]["model_type_counters"] == {
                    "<class 'sample_lib.modules.sample_task.sample_implementation.SampleModule'>": 1
                }


def test_reflection_enabled(runtime_grpc_server):
    """This pings the reflection API to ensure we can list the caikit services that are running"""
    # See https://github.com/grpc/grpc/blob/master/doc/python/server_reflection.md
    channel = runtime_grpc_server.make_local_channel()
    reflection_db = ProtoReflectionDescriptorDatabase(channel)

    desc_pool = DescriptorPool(reflection_db)
    service_desc = desc_pool.FindServiceByName(
        "caikit.runtime.SampleLib.SampleLibService"
    )
    method_desc = service_desc.FindMethodByName("SampleTaskPredict")
    assert method_desc is not None


def test_streaming_responses_work(runtime_grpc_server):
    """This test uses the health check's Watch rpc to ensure that a unary->stream RPC functions
    as expected."""
    stub = health_pb2_grpc.HealthStub(runtime_grpc_server.make_local_channel())
    req = health_pb2.HealthCheckRequest()
    for response in stub.Watch(req):
        assert response is not None
        break


def test_streaming_handlers_are_built_correctly(runtime_grpc_server):
    """This is a very cheat-y test of a private method to check that we build our internal
    handlers using the correct handler function"""

    class FakeHandler:
        pass

    # NB: the unary_stream case is tested via health check watch in `test_streaming_responses_work`
    # (unary_unary cases are checked in every other test)

    stream_unary_handler = RpcMethodHandler(
        request_streaming=True,
        response_streaming=False,
        request_deserializer="foo",
        response_serializer="bar",
        unary_unary=None,
        stream_unary=FakeHandler,
        unary_stream=None,
        stream_stream=None,
    )
    new_handler = runtime_grpc_server.server._make_new_handler(stream_unary_handler)
    assert new_handler.stream_unary is not None
    assert new_handler.stream_unary.__name__ == "safe_rpc_call"

    stream_stream_handler = RpcMethodHandler(
        request_streaming=True,
        response_streaming=True,
        request_deserializer="foo",
        response_serializer="bar",
        unary_unary=None,
        stream_unary=None,
        unary_stream=None,
        stream_stream=FakeHandler,
    )
    new_handler = runtime_grpc_server.server._make_new_handler(stream_stream_handler)
    assert new_handler.stream_stream is not None
    assert new_handler.stream_stream.__name__ == "safe_rpc_call"


def test_grpc_sever_shutdown_with_model_poll(open_port):
    """Test that a SIGINT successfully shuts down the running server"""
    with tempfile.TemporaryDirectory() as workdir:
        server_proc = ModuleSubproc(
            "caikit.runtime.grpc_server",
            kill_timeout=30.0,
            RUNTIME_GRPC_PORT=str(open_port),
            RUNTIME_LOCAL_MODELS_DIR=workdir,
            RUNTIME_LAZY_LOAD_LOCAL_MODELS="true",
            RUNTIME_LAZY_LOAD_POLL_PERIOD_SECONDS="0.1",
            RUNTIME_METRICS_ENABLED="false",
        )
        with server_proc as proc:
            # Wait for the server to be up
            _assert_connection(
                grpc.insecure_channel(f"localhost:{open_port}"), max_failures=500
            )

            # Signal the server to shut down
            proc.send_signal(signal.SIGINT)

        # Make sure the process was not killed
        assert not server_proc.killed


def test_construct_with_options(open_port, sample_train_service, sample_int_file):
    """Make sure that the server can be booted with config options"""
    with temp_config(
        {
            "runtime": {
                "grpc": {
                    "port": open_port,
                    "options": {
                        "grpc.max_receive_message_length": 1,
                    },
                }
            }
        },
        "merge",
    ):
        server = RuntimeGRPCServer()
        with server:
            # Send a request with a payload that's too big and make sure it gets
            # rejected
            stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
            training_data = stream_type(
                file=stream_type.File(filename=sample_int_file)
            ).to_proto()
            train_request = sample_train_service.messages.OtherTaskOtherModuleTrainRequest(
                model_name="Bar Training",
                parameters=sample_train_service.messages.OtherTaskOtherModuleTrainParameters(
                    sample_input_sampleinputtype=SampleInputType(
                        name="Gabe"
                    ).to_proto(),
                    batch_size=100,
                    training_data=training_data,
                ),
            )
            train_stub = sample_train_service.stub_class(server.make_local_channel())
            with pytest.raises(grpc.RpcError) as context:
                train_stub.OtherTaskOtherModuleTrain(train_request)
            assert context.value.code() == grpc.StatusCode.RESOURCE_EXHAUSTED


# Test implementation details #########################
@dataclass
class KeyPair:
    cert: str
    key: str


@dataclass
class TLSConfig:
    server: KeyPair
    client: KeyPair


def _make_secure_channel(
    server: RuntimeGRPCServer,
    ca_cert: str,
    client_key: str = None,
    client_cert: str = None,
):
    if client_key and client_cert:
        # mTLS
        credentials = grpc.ssl_channel_credentials(
            root_certificates=bytes(ca_cert, "utf-8"),
            private_key=bytes(client_key, "utf-8"),
            certificate_chain=bytes(client_cert, "utf-8"),
        )
    else:
        # TLS
        credentials = grpc.ssl_channel_credentials(
            root_certificates=bytes(ca_cert, "utf-8")
        )
    return grpc.secure_channel(f"localhost:{server.port}", credentials=credentials)


def _assert_connection(channel, max_failures=20):
    """Check that we can ping the server on this channel.

    Assumes that it will come up, but it's hard to distinguish between a failure to boot and a
    failure of TLS communication.
    """
    done = False
    failures = 0
    while not done:
        try:
            health_pb2_grpc.HealthStub(channel).Check(health_pb2.HealthCheckRequest())
            done = True
        except grpc.RpcError as e:
            log.debug(
                f"[RpcError] {e}; will try to reconnect to test server in 0.1 second."
            )

            time.sleep(0.1)
            failures += 1
            if failures > max_failures:
                raise e



## Tests #######################################################################
# Standard
from typing import Iterable, Union
import uuid

# Third Party
import pytest

# Local
from caikit.core import TaskBase, task
from sample_lib import SampleModule
from sample_lib.data_model.sample import SampleInputType, SampleOutputType, SampleTask
import caikit.core


def test_task_decorator_has_streaming_types():
    @task(
        unary_parameters={"text": str},
        streaming_parameters={"tokens": Iterable[str]},
        unary_output_type=SampleOutputType,
        streaming_output_type=Iterable[SampleOutputType],
    )
    class SampleTask(TaskBase):
        pass

    assert (
        SampleTask.get_output_type(output_streaming=True) == Iterable[SampleOutputType]
    )
    assert SampleTask.get_required_parameters(input_streaming=True) == {
        "tokens": Iterable[str]
    }


def test_task_decorator_validates_class_extends_task_base():
    with pytest.raises(TypeError, match="is not a subclass of .*TaskBase"):

        @task()
        class SampleTask:
            pass


def test_task_decorator_validates_output_is_data_model():
    with pytest.raises(TypeError, match=".*str.* is not a subclass of .*DataBase"):

        @task(unary_parameters={"text": str}, unary_output_type=str)
        class SampleTask(TaskBase):
            pass


def test_task_decorator_can_have_iterable_output():
    """This test covers tasks + modules with streaming output"""

    @task(
        unary_parameters={"sample_input": SampleInputType},
        streaming_output_type=Iterable[SampleOutputType],
    )
    class StreamingTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()),
        name="StreamingModule",
        version="0.0.1",
        task=StreamingTask,
    )
    class StreamingModule(caikit.core.ModuleBase):
        @StreamingTask.taskmethod(output_streaming=True)
        def run(
            self, sample_input: SampleInputType
        ) -> caikit.core.data_model.DataStream[SampleOutputType]:
            pass


def test_task_decorator_validates_streaming_output_is_iterable():
    with pytest.raises(TypeError, match="not a subclass of .*Iterable"):

        @task(streaming_parameters={"text": Iterable[str]}, streaming_output_type=str)
        class StreamingTask(TaskBase):
            pass


def test_task_decorator_validates_streaming_input_is_iterable():
    with pytest.raises(TypeError, match="not a subclass of .*Iterable"):

        @task(
            streaming_parameters={"text": str},
            streaming_output_type=Iterable[SampleOutputType],
        )
        class StreamingTask(TaskBase):
            pass


def test_task_validator_raises_on_wrong_streaming_type():
    @task(
        unary_parameters={"sample_input": SampleInputType},
        streaming_output_type=Iterable[SampleOutputType],
    )
    class StreamingTask(TaskBase):
        pass

    with pytest.raises(TypeError, match="Wrong output type for module"):

        @caikit.core.module(
            id=str(uuid.uuid4()),
            name="InvalidStreamingModule",
            version="0.0.1",
            task=StreamingTask,
        )
        class InvalidStreamingModule(caikit.core.ModuleBase):
            @StreamingTask.taskmethod(output_streaming=True)
            def run(
                self, sample_input: SampleInputType
            ) -> caikit.core.data_model.DataStream[SampleInputType]:
                pass


def test_task_is_set_on_module_classes():
    assert hasattr(SampleModule, "TASK_CLASS")
    assert SampleModule.TASK_CLASS == SampleTask


def test_task_can_be_inferred_from_parent_module():
    @caikit.core.modules.module(id="foobar", name="Stuff", version="0.0.1")
    class Stuff(SampleModule):
        pass

    assert Stuff.TASK_CLASS == SampleModule.TASK_CLASS


def test_task_cannot_conflict_with_parent_module():
    @task(unary_parameters={"foo": SampleInputType}, unary_output_type=SampleOutputType)
    class SomeTask(TaskBase):
        pass

    with pytest.raises(TypeError, match="but superclass has"):

        @caikit.core.modules.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(SampleModule):
            pass


def test_task_is_not_required_for_modules():
    @caikit.core.modules.module(id=str(uuid.uuid4()), name="Stuff", version="0.0.1")
    class Stuff(caikit.core.ModuleBase):
        pass

    assert Stuff.TASK_CLASS is None


def test_task_validation_throws_on_no_params():
    @task(unary_parameters={"foo": int}, unary_output_type=SampleOutputType)
    class SomeTask(TaskBase):
        pass

    with pytest.raises(
        ValueError,
        match="Task could not be validated, no .run parameters were provided",
    ):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(caikit.core.ModuleBase):
            def run(self) -> SampleOutputType:
                pass


def test_task_validation_throws_on_no_return_type():
    @task(unary_parameters={"foo": int}, unary_output_type=SampleOutputType)
    class SomeTask(TaskBase):
        pass

    with pytest.raises(
        ValueError,
        match="Task could not be validated, no .run return type was provided",
    ):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(caikit.core.ModuleBase):
            def run(self, foo: int):
                pass


def test_task_validation_throws_on_wrong_return_type():
    @task(
        unary_parameters={"foo": int},
        unary_output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    with pytest.raises(
        TypeError,
        match="Wrong output type for module",
    ):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(caikit.core.ModuleBase):
            def run(self, foo: int) -> SampleInputType:
                pass


def test_task_validation_accepts_union_outputs():
    @task(
        unary_parameters={"foo": int},
        unary_output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
    )
    class Stuff(caikit.core.ModuleBase):
        def run(self, foo: int) -> Union[SampleOutputType, int, str]:
            pass


def test_task_validation_throws_on_missing_parameter():
    @task(
        unary_parameters={"foo": int},
        unary_output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    with pytest.raises(TypeError, match="Required parameters .*foo.* not in signature"):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(caikit.core.ModuleBase):
            def run(self, bar: str) -> SampleOutputType:
                pass


def test_task_validation_throws_on_wrong_parameter_type():
    @task(
        unary_parameters={"foo": int},
        unary_output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    with pytest.raises(
        TypeError,
        match="Parameter foo has type .*str.* but type .*int.* is required",
    ):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class Stuff(caikit.core.ModuleBase):
            def run(self, foo: str) -> SampleOutputType:
                pass


def test_task_validation_passes_when_module_has_correct_run_signature():
    @task(
        unary_parameters={"foo": int},
        unary_output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
    )
    class SomeModule(caikit.core.ModuleBase):
        def run(self, foo: int, bar: str) -> SampleOutputType:
            pass


def test_task_decorator_adds_taskmethods_to_modules():
    """This test covers tasks + modules with streaming output"""

    @task(
        unary_parameters={"sample_input": SampleInputType},
        unary_output_type=SampleOutputType,
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        streaming_output_type=Iterable[SampleOutputType],
    )
    class StreamingTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()),
        name="StreamingModule",
        version="0.0.1",
        task=StreamingTask,
    )
    class StreamingModule(caikit.core.ModuleBase):
        @StreamingTask.taskmethod()
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

        @StreamingTask.taskmethod(output_streaming=True)
        def run_stream_out(
            self, sample_input: SampleInputType
        ) -> caikit.core.data_model.DataStream[SampleOutputType]:
            pass

        @StreamingTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: caikit.core.data_model.DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

    assert (
        StreamingModule.get_inference_signature(
            input_streaming=False, output_streaming=False
        ).method_name
        == "run"
    )
    assert (
        StreamingModule.get_inference_signature(
            input_streaming=False, output_streaming=True
        ).method_name
        == "run_stream_out"
    )

    assert (
        StreamingModule.get_inference_signature(
            input_streaming=True, output_streaming=False
        ).method_name
        == "run_stream_in"
    )


def test_task_decorator_datastream_throws_wrong_type():
    @task(
        unary_parameters={"sample_input": SampleInputType},
        unary_output_type=SampleOutputType,
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        streaming_output_type=Iterable[SampleOutputType],
    )
    class DataStreamStreamingTask(TaskBase):
        pass

    with pytest.raises(
        TypeError,
        match=".*expected .*SampleInputType.* but got .*SampleOutputType",
    ):

        @caikit.core.module(
            id=str(uuid.uuid4()),
            name="DataStreamStreamingModule",
            version="0.0.1",
            task=DataStreamStreamingTask,
        )
        class DataStreamStreamingModule(caikit.core.ModuleBase):
            @DataStreamStreamingTask.taskmethod(input_streaming=True)
            def run_stream_in(
                self, sample_inputs: caikit.core.data_model.DataStream[SampleOutputType]
            ) -> SampleOutputType:
                pass


def test_task_decorator_datastream_params():
    @task(
        unary_parameters={"sample_input": SampleInputType},
        unary_output_type=SampleOutputType,
        streaming_parameters={"sample_inputs": Iterable[SampleInputType]},
        streaming_output_type=Iterable[SampleOutputType],
    )
    class DataStreamStreamingTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()),
        name="DataStreamStreamingModule",
        version="0.0.1",
        task=DataStreamStreamingTask,
    )
    class DataStreamStreamingModule(caikit.core.ModuleBase):
        @DataStreamStreamingTask.taskmethod()
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

        @DataStreamStreamingTask.taskmethod(output_streaming=True)
        def run_stream_out(
            self, sample_input: SampleInputType
        ) -> caikit.core.data_model.DataStream[SampleOutputType]:
            pass

        @DataStreamStreamingTask.taskmethod(input_streaming=True)
        def run_stream_in(
            self, sample_inputs: caikit.core.data_model.DataStream[SampleInputType]
        ) -> SampleOutputType:
            pass

        @DataStreamStreamingTask.taskmethod(input_streaming=True, output_streaming=True)
        def run_stream_bidi(
            self, sample_inputs: caikit.core.data_model.DataStream[SampleInputType]
        ) -> caikit.core.data_model.DataStream[SampleOutputType]:
            pass

    stream_stream_method_signature = DataStreamStreamingModule.get_inference_signature(
        input_streaming=False, output_streaming=False
    )

    out_stream_method_signature = DataStreamStreamingModule.get_inference_signature(
        input_streaming=False, output_streaming=True
    )

    in_stream_method_signature = DataStreamStreamingModule.get_inference_signature(
        input_streaming=True, output_streaming=False
    )

    bidi_stream_method_signature = DataStreamStreamingModule.get_inference_signature(
        input_streaming=True, output_streaming=True
    )

    assert stream_stream_method_signature.method_name == "run"
    assert stream_stream_method_signature.parameters == {
        "sample_input": SampleInputType
    }
    assert stream_stream_method_signature.return_type == SampleOutputType

    assert out_stream_method_signature.method_name == "run_stream_out"
    assert out_stream_method_signature.parameters == {"sample_input": SampleInputType}
    assert (
        out_stream_method_signature.return_type
        == caikit.core.data_model.streams.data_stream.DataStream[SampleOutputType]
    )

    assert in_stream_method_signature.method_name == "run_stream_in"
    assert in_stream_method_signature.parameters == {
        "sample_inputs": caikit.core.data_model.streams.data_stream.DataStream[
            SampleInputType
        ]
    }
    assert in_stream_method_signature.return_type == SampleOutputType

    assert bidi_stream_method_signature.method_name == "run_stream_bidi"
    assert bidi_stream_method_signature.parameters == {
        "sample_inputs": caikit.core.data_model.streams.data_stream.DataStream[
            SampleInputType
        ]
    }
    assert (
        bidi_stream_method_signature.return_type
        == caikit.core.data_model.streams.data_stream.DataStream[SampleOutputType]
    )


def test_decorator_adds_default_run_method_to_modules():
    @task(
        unary_parameters={"sample_input": SampleInputType},
        unary_output_type=SampleOutputType,
        streaming_output_type=Iterable[SampleOutputType],
    )
    class SomeTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()),
        name="SomeModule",
        version="0.0.1",
        task=SomeTask,
    )
    class SomeModule(caikit.core.ModuleBase):
        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            pass

    assert (
        SomeModule.get_inference_signature(
            input_streaming=False, output_streaming=False
        ).method_name
        == "run"
    )


# ----------- BACKWARDS COMPATIBILITY ------------------------------------------- ##


def test_task_backwards_compatibility():
    """The old 'required_parameters' and 'output_type' should continue to function"""

    @task(
        required_parameters={"foo": int},
        output_type=SampleOutputType,
    )
    class SomeTask(TaskBase):
        pass

    @caikit.core.module(
        id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
    )
    class SomeModule(caikit.core.ModuleBase):
        def run(self, foo: int, bar: str) -> SampleOutputType:
            pass

    with pytest.raises(TypeError):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class SomeModule(caikit.core.ModuleBase):
            def run(self, foo: str) -> SampleOutputType:
                pass

    with pytest.raises(TypeError):

        @caikit.core.module(
            id=str(uuid.uuid4()), name="Stuff", version="0.0.1", task=SomeTask
        )
        class SomeModule(caikit.core.ModuleBase):
            def run(self, foo: int) -> SampleInputType:
                pass



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.core.data_model import DataStream
from caikit.core.data_model.streams.converter import DataStreamConverter

# Unit Test Infrastructure
from tests.base import TestCaseBase

LIST_OF_DICTS = [
    {"a": 1, "b": 2, "c": 3},
    {"a": 4, "b": 5, "c": 6},
    {"a": 7, "b": 8, "c": 9},
]

LIST_OF_LISTS = [
    [1, 3],
    [4, 6],
    [7, 9],
]


class TestDataStreamConverter(TestCaseBase):
    def test_errors_with_unsupported_target_type(self):
        with self.assertRaises(ValueError):
            DataStreamConverter(tuple, ["a", "c"])

    def test_it_is_okay_with_empty_streams_to_list(self):
        empty_stream = DataStream.from_iterable([])
        to_list_converter = DataStreamConverter(list, ["a", "c"])
        stream = to_list_converter.convert(empty_stream)
        self.assertEqual([], list(stream))

    def test_it_is_okay_with_empty_streams_to_dict(self):
        empty_stream = DataStream.from_iterable([])
        to_dict_converter = DataStreamConverter(dict, ["a", "c"])
        stream = to_dict_converter.convert(empty_stream)
        self.assertEqual([], list(stream))

    def test_can_convert_dict_stream_to_list_stream(self):
        dict_stream = DataStream.from_iterable(LIST_OF_DICTS)
        converter = DataStreamConverter(list, ["a", "c"])
        list_stream = converter.convert(dict_stream)
        expected_lists = [
            [1, 3],
            [4, 6],
            [7, 9],
        ]
        self.assertEqual(expected_lists, list(list_stream))

    def test_can_convert_list_stream_to_dict_stream(self):
        list_stream = DataStream.from_iterable(LIST_OF_LISTS)
        converter = DataStreamConverter(dict, ["a", "b"])
        dict_stream = converter.convert(list_stream)
        expected_dict = [
            {"a": 1, "b": 3},
            {"a": 4, "b": 6},
            {"a": 7, "b": 9},
        ]
        self.assertEqual(expected_dict, list(dict_stream))

    def test_does_not_convert_list_stream_already_of_target_type(self):
        # this only actually works if the input and target lists are the same length
        list_stream = DataStream.from_iterable(LIST_OF_LISTS)
        converter = DataStreamConverter(list, ["1", "3"])
        converted_stream = converter.convert(list_stream)
        self.assertEqual(list(converted_stream), list(list_stream))

    def test_does_not_convert_dict_stream_already_of_target_type(self):
        # this only actually works if the input dicts have all the expected keys
        dict_stream = DataStream.from_iterable(LIST_OF_DICTS)
        converter = DataStreamConverter(dict, ["b", "a", "c"])
        converted_stream = converter.convert(dict_stream)
        self.assertEqual(list(converted_stream), list(dict_stream))



"""
A sample module for sample things!
"""
# Standard
from typing import List, Optional

# Local
from ...data_model.sample import (
    SampleInputType,
    SampleOutputType,
    SampleTask,
    SampleTrainingType,
)
from caikit.core.data_model import DataStream
from caikit.core.modules import ModuleLoader, ModuleSaver
import caikit.core


@caikit.core.module(
    "00af2203-0405-0607-0263-0a0b02dd0c2f", "ListModule", "0.0.1", SampleTask
)
class ListModule(caikit.core.ModuleBase):
    POISON_PILL_NAME = "Bob Marley"

    def __init__(self, batch_size=64, learning_rate=0.0015):
        super().__init__()
        self.batch_size = batch_size
        self.learning_rate = learning_rate

    @classmethod
    def load(cls, model_path, **kwargs):
        loader = ModuleLoader(model_path)
        config = loader.config
        return cls(config["train"]["batch_size"], config["train"]["learning_rate"])

    def run(
        self, sample_input: SampleInputType, throw: bool = False
    ) -> SampleOutputType:
        """
        Args:
            sample_input (sample_lib.data_model.SampleInputType): the input

        Returns:
            sample_lib.data_model.SampleOutputType: The output
        """
        if throw:
            raise RuntimeError("barf!")
        if sample_input.name == self.POISON_PILL_NAME:
            raise ValueError(f"{self.POISON_PILL_NAME} is not allowed!")
        return SampleOutputType(f"Hello {sample_input.name}")

    def save(self, model_path):
        module_saver = ModuleSaver(
            self,
            model_path=model_path,
        )
        with module_saver:
            config_options = {
                "train": {
                    "batch_size": self.batch_size,
                    "learning_rate": self.learning_rate,
                },
            }

            module_saver.update_config(config_options)

    @classmethod
    def train(
        cls,
        training_data: DataStream[SampleTrainingType],
        batch_size: int = 64,
        poison_pills: Optional[List[str]] = None,
    ) -> "ListModule":
        """Sample training method that produces a trained model"""
        # Barf if we were incorrectly passed data not in datastream format
        poison_pills = (
            poison_pills if poison_pills is not None else [cls.POISON_PILL_NAME]
        )
        assert isinstance(training_data, DataStream)
        assert len(poison_pills) > 0
        return cls(batch_size=batch_size)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import os
import tempfile

# Third Party
import pytest

# Local
from caikit.core.data_model.base import DataBase
from caikit.runtime.protobufs import model_runtime_pb2
from caikit.runtime.service_generation.data_stream_source import DataStreamSourceBase
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import get_data_model
from caikit.runtime.utils.servicer_util import (
    build_caikit_library_request_dict,
    build_proto_response,
    get_metadata,
    is_protobuf_primitive_field,
    validate_caikit_library_class_exists,
    validate_caikit_library_class_method_exists,
    validate_data_model,
)
from sample_lib import SamplePrimitiveModule
from sample_lib.data_model import SampleInputType
from tests.conftest import random_test_id
from tests.fixtures import Fixtures
import caikit.core
import caikit.interfaces
import sample_lib


# ---------------- Tests for validate_caikit_library_class_exists --------------------
def test_servicer_util_validate_caikit_library_class_exists_returns_caikit_class():
    """Test that validate_caikit_library_class_exists returns a caikit class object"""
    validated_class = validate_caikit_library_class_exists(
        get_data_model(), "TrainingStatusResponse"
    )

    assert issubclass(
        validated_class,
        caikit.interfaces.runtime.data_model.training_management.TrainingStatusResponse,
    )

    """
    We really wanted to avoid having concrete data model objects live in caikit.core,
    but because ModuleBase used ProducerId, we had to keep that there. That said,
    it is forwarded to caikit.interfaces.common, so from a user's perspective,
    we should consider it to be part of caikit.interfaces.common not
    caikit.core.data_model."""

    validated_class = validate_caikit_library_class_exists(
        get_data_model(), "ProducerId"
    )
    assert validated_class == caikit.interfaces.common.data_model.producer.ProducerId


def test_servicer_util_validate_caikit_library_class_exists_raises_for_garbage_input():
    """Test that validate_caikit_library_class_exists raises for garbage 'model'"""
    with pytest.raises(ValueError):
        validate_caikit_library_class_exists(get_data_model(), "NonExistentClass")


# ---------------- Tests for validate_caikit_library_class_method_exists --------------------
def test_service_util_validate_caikit_library_class_method_exists_doesnt_raise():
    """Test that validate_caikit_library_class_method_exists doesn't raise for valid methods"""
    try:
        validate_caikit_library_class_method_exists(
            caikit.interfaces.runtime.data_model.training_management.ModelPointer,
            "to_proto",
        )
        validate_caikit_library_class_method_exists(
            caikit.interfaces.runtime.data_model.training_management.ModelPointer,
            "from_proto",
        )
    except Exception as e:
        assert (
            False
        ), "validate_caikit_library_class_method_exists failed for ModelPointer!"

    try:
        validate_caikit_library_class_method_exists(
            caikit.interfaces.common.data_model.producer.ProducerId, "to_proto"
        )
        validate_caikit_library_class_method_exists(
            caikit.interfaces.common.data_model.producer.ProducerId, "from_proto"
        )
    except Exception as e:
        assert False, "validate_caikit_library_class_method_exists failed!"


def test_service_util_validate_caikit_library_class_method_exists_does_raise():
    """Test that validate_caikit_library_class_method_exists raises for garbage input method"""
    with pytest.raises(AttributeError):
        validate_caikit_library_class_method_exists(
            caikit.interfaces.runtime.data_model.ModelPointer, "non_existent_method"
        )


# ---------------- Tests for build_proto_response  --------------------------
def test_servicer_util_build_proto_response_raises_on_garbage_response_type():
    class FooResponse:
        def __init__(self, foo) -> None:
            self.foo = foo

    with pytest.raises(CaikitRuntimeException):
        build_proto_response(FooResponse(foo="This is a foo response"))


# ---------------- Tests for is_protobuf_primitive_field --------------------
def test_servicer_util_is_protobuf_primitive_returns_true_for_primitive_types(
    sample_inference_service,
):
    """Test that is_protobuf_primitive_field is True when considering primitive types"""
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    assert is_protobuf_primitive_field(
        predict_class().to_proto().DESCRIPTOR.fields_by_name["int_type"]
    )


def test_servicer_util_is_protobuf_primitive_returns_false_for_custom_types(
    sample_inference_service,
):
    """Test that is_protobuf_primitive_field is False when considering message and
    enum types. This is essential for handling Caikit library CDM objects, which are
    generally defined in terms of messages"""
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    assert not is_protobuf_primitive_field(
        predict_class().to_proto().DESCRIPTOR.fields_by_name["sample_input"]
    )


def test_servicer_util_is_protobuf_primitive_returns_false_for_instance_not_in_FieldDescriptor():
    field = caikit.interfaces.runtime.data_model.ModelPointer
    assert False == (is_protobuf_primitive_field(field))


# ---------------- Tests for get_metadata --------------------
def test_servicer_util_get_metadata_raises_on_no_metadata_in_context_but_key_required():
    """Test that get_metadata raises on no metadata context, but the key is required"""
    with pytest.raises(CaikitRuntimeException):
        get_metadata(context=None, key="some-key", required=True)


def test_servicer_util_get_metadata_returns_none_for_absent_key():
    """Test that get_metadata returns None for absent invocation metadata key"""
    assert (
        get_metadata(
            Fixtures.build_context("syntax_izumo_v0-0-1_en"), "some-missing-key"
        )
        == None
    )


def test_servicer_util_get_metadata_should_get_mmmodelid():
    """Test that get_metadata returns value for present invocation metadata key"""
    assert (
        get_metadata(Fixtures.build_context("syntax_izumo_v0-0-1_en"), "mm-model-id")
        == "syntax_izumo_v0-0-1_en"
    )


# ---------------- Tests for validate_data_model --------------------
def test_servicer_util_validates_caikit_core_data_model(
    sample_inference_service, sample_train_service
):
    """Test that validate_data_model validates Caikit library CDM on initialization"""

    try:
        validate_data_model(sample_train_service.descriptor)
    except Exception as e:
        assert False, "Validation of interfaces failed for train!"

    try:
        validate_data_model(sample_inference_service.descriptor)
    except Exception as e:
        assert False, "Validation of interfaces failed for Predict!"


def test_servicer_util_will_not_validate_arbitrary_service_descriptor():
    """Test that validate_data_model raises exception validating arbitrary ServiceDescriptor"""
    with pytest.raises(ValueError):
        validate_data_model(model_runtime_pb2._MODELRUNTIME)


# ---------------- Tests for build_caikit_library_request_dict --------------------
HAPPY_PATH_INPUT_DM = SampleInputType(name="Gabe")


def test_global_predict_build_caikit_library_request_dict_creates_caikit_core_run_kwargs(
    sample_inference_service,
):
    """Test that build_caikit_library_request_dict creates module run kwargs from RPC msg"""
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    request_dict = build_caikit_library_request_dict(
        predict_class(sample_input=HAPPY_PATH_INPUT_DM).to_proto(),
        sample_lib.modules.sample_task.SampleModule.RUN_SIGNATURE,
    )

    # Since using pythonic data model and throw has a default parameter, it is an expected argument
    expected_arguments = {"sample_input", "throw"}

    assert expected_arguments == set(request_dict.keys())
    assert isinstance(request_dict["sample_input"], SampleInputType)


def test_global_predict_build_caikit_library_request_dict_strips_invalid_run_kwargs_from_request(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict strips invalid run kwargs from request"""
    # Sample module doesn't take the `int_type` or `bool_type` params
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    request_dict = build_caikit_library_request_dict(
        predict_class(
            sample_input=HAPPY_PATH_INPUT_DM,
            int_type=5,
            bool_type=True,
        ).to_proto(),
        sample_lib.modules.sample_task.SampleModule.RUN_SIGNATURE,
    )

    expected_arguments = {"sample_input", "throw"}
    assert expected_arguments == set(request_dict.keys())
    assert "int_type" not in request_dict.keys()


def test_global_predict_build_caikit_library_request_dict_strips_empty_list_from_request(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict strips empty list from request"""
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    request_dict = build_caikit_library_request_dict(
        predict_class(int_type=5, list_type=[]).to_proto(),
        sample_lib.modules.sample_task.SamplePrimitiveModule.RUN_SIGNATURE,
    )

    assert "list_type" not in request_dict.keys()
    assert "int_type" in request_dict.keys()


def test_global_predict_build_caikit_library_request_dict_with_proto_does_not_include_unset_primitives(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict works for primitives"""
    # When using protobuf message for request, if params unset, does not include primitive args
    # that are unset even if default values present
    request = sample_inference_service.messages.SampleTaskRequest()

    request_dict = build_caikit_library_request_dict(
        request, SamplePrimitiveModule.RUN_SIGNATURE
    )
    # None of the primitive args should be there
    assert len(request_dict.keys()) == 0


def test_global_predict_build_caikit_library_request_dict_works_for_unset_primitives(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict works for primitives"""
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    request = predict_class().to_proto()

    request_dict = build_caikit_library_request_dict(
        request, SamplePrimitiveModule.RUN_SIGNATURE
    )
    # When using pythonic data model for request, primitive args found
    # because default values set
    assert len(request_dict.keys()) == 5
    assert request_dict["bool_type"] is True
    assert request_dict["int_type"] == 42
    assert request_dict["float_type"] == 34.0
    assert request_dict["str_type"] == "moose"
    assert request_dict["bytes_type"] == b""


def test_global_predict_build_caikit_library_request_dict_with_proto_for_set_primitives(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict works for primitives"""
    # When using protobuf message for request, only set params included
    request = sample_inference_service.messages.SampleTaskRequest(
        int_type=5,
        float_type=4.2,
        str_type="moose",
        bytes_type=b"foo",
        list_type=["1", "2", "3"],
    )

    request_dict = build_caikit_library_request_dict(
        request, SamplePrimitiveModule.RUN_SIGNATURE
    )
    assert request_dict["int_type"] == 5
    assert request_dict["float_type"] == 4.2
    assert request_dict["str_type"] == "moose"
    assert request_dict["bytes_type"] == b"foo"
    assert request_dict["list_type"] == ["1", "2", "3"]


def test_global_predict_build_caikit_library_request_dict_works_for_set_primitives(
    sample_inference_service,
):
    """Global predict build_caikit_library_request_dict works for primitives"""
    # When using pythonic data model for request, primitive args found that are set and
    # ones with default values set
    predict_class = DataBase.get_class_for_name("SampleTaskRequest")
    request = predict_class(
        int_type=5,
        float_type=4.2,
        str_type="moose",
        bytes_type=b"foo",
        list_type=["1", "2", "3"],
    ).to_proto()

    request_dict = build_caikit_library_request_dict(
        request, SamplePrimitiveModule.RUN_SIGNATURE
    )
    # bool_type also included because default value set
    assert request_dict["bool_type"] is True
    assert request_dict["int_type"] == 5
    assert request_dict["float_type"] == 4.2
    assert request_dict["str_type"] == "moose"
    assert request_dict["bytes_type"] == b"foo"
    assert request_dict["list_type"] == ["1", "2", "3"]


@pytest.mark.skip(
    "Skipping until bug fixes for unset Union fields - https://github.com/caikit/caikit/issues/471"
)
def test_global_train_build_caikit_library_request_dict_strips_empty_list_from_request(
    sample_train_service,
):
    """Global train build_caikit_library_request_dict strips empty list from request"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(jsondata=stream_type.JsonData(data=[]))
    train_class = DataBase.get_class_for_name("SampleTaskSampleModuleTrainRequest")
    train_request_params_class = DataBase.get_class_for_name(
        "SampleTaskSampleModuleTrainParameters"
    )
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(training_data=training_data),
    ).to_proto()

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.sample_task.SampleModule.TRAIN_SIGNATURE,
    )

    # model_name is not expected to be passed through
    # since using pythonic data model, keeps params that have default value and removes ones that are empty
    expected_arguments = {
        "sleep_time",
        "oom_exit",
        "sleep_increment",
        "batch_size",
    }

    assert expected_arguments == set(caikit.core_request.keys())


@pytest.mark.skip(
    "Skipping until bug fixes for unset Union fields - https://github.com/caikit/caikit/issues/471"
)
def test_global_train_build_caikit_library_request_dict_with_proto_keeps_empty_params_from_request(
    sample_train_service,
):
    """Global train build_caikit_library_request_dict strips empty list from request"""
    # NOTE: Using protobuf to create request, by explicitly passing in training_data, even though
    # the datastream is empty, it's not removed from request, which is expected
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(jsondata=stream_type.JsonData(data=[])).to_proto()
    train_request = sample_train_service.messages.SampleTaskSampleModuleTrainRequest(
        model_name=random_test_id(),
        parameters=sample_train_service.messages.SampleTaskSampleModuleTrainParameters(
            training_data=training_data
        ),
    )

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.sample_task.SampleModule.TRAIN_SIGNATURE,
    )

    # model_name is not expected to be passed through
    expected_arguments = {"training_data"}

    assert expected_arguments == set(caikit.core_request.keys())
    assert isinstance(caikit.core_request["training_data"], DataStreamSourceBase)


def test_global_train_build_caikit_library_request_dict_works_for_repeated_fields(
    sample_train_service,
):
    """Global train build_caikit_library_request_dict works for repeated fields"""

    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(jsondata=stream_type.JsonData(data=[]))
    train_class = DataBase.get_class_for_name("SampleTaskListModuleTrainRequest")
    train_request_params_class = DataBase.get_class_for_name(
        "SampleTaskListModuleTrainParameters"
    )
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            training_data=training_data,
            poison_pills=["Bob Marley", "Bunny Livingston"],
        ),
    ).to_proto()

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.sample_task.ListModule.TRAIN_SIGNATURE,
    )

    # model_name is not expected to be passed through
    expected_arguments = {"poison_pills", "batch_size"}

    assert expected_arguments == set(caikit.core_request.keys())
    assert len(caikit.core_request.keys()) == 2
    assert "poison_pills" in caikit.core_request
    assert isinstance(caikit.core_request["poison_pills"], list)


def test_global_train_build_caikit_library_request_dict_ok_with_DataStreamSourceInt_with_inline_json(
    sample_train_service,
):
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceInt
    training_data = stream_type(jsondata=stream_type.JsonData(data=[100, 120]))

    train_class = DataBase.get_class_for_name("OtherTaskOtherModuleTrainRequest")
    train_request_params_class = DataBase.get_class_for_name(
        "OtherTaskOtherModuleTrainParameters"
    )
    train_request = train_class(
        model_name="Bar Training",
        parameters=train_request_params_class(
            sample_input_sampleinputtype=HAPPY_PATH_INPUT_DM,
            batch_size=100,
            training_data=training_data,
        ),
    ).to_proto()

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.other_task.OtherModule.TRAIN_SIGNATURE,
    )

    expected_arguments = {"batch_size", "sample_input", "training_data"}

    assert expected_arguments == set(caikit.core_request.keys())


@pytest.mark.skip(
    "Skipping until bug fixes for unset Union fields - https://github.com/caikit/caikit/issues/471"
)
def test_global_train_build_caikit_library_request_dict_ok_with_data_stream_file_type_csv(
    sample_train_service, sample_csv_file
):
    """Global train build_caikit_library_request_dict works for csv training data file"""

    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(file=stream_type.File(filename=sample_csv_file))
    train_class = DataBase.get_class_for_name("SampleTaskSampleModuleTrainRequest")
    train_request_params_class = DataBase.get_class_for_name(
        "SampleTaskSampleModuleTrainParameters"
    )
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            training_data=training_data,
        ),
    ).to_proto()

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.sample_task.SampleModule.TRAIN_SIGNATURE,
    )

    # model_name is not expected to be passed through
    expected_arguments = {
        "training_data",
        "batch_size",
        "oom_exit",
        "sleep_time",
        "sleep_increment",
    }

    assert expected_arguments == set(caikit.core_request.keys())


def test_global_train_build_caikit_library_request_dict_ok_with_training_data_as_list_of_files(
    sample_train_service, sample_csv_file, sample_json_file
):
    """Global train build_caikit_library_request_dict works for list of data files"""
    stream_type = caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
    training_data = stream_type(
        listoffiles=stream_type.ListOfFiles(files=[sample_csv_file, sample_json_file])
    )
    train_class = DataBase.get_class_for_name("SampleTaskListModuleTrainRequest")
    train_request_params_class = DataBase.get_class_for_name(
        "SampleTaskListModuleTrainParameters"
    )
    train_request = train_class(
        model_name=random_test_id(),
        parameters=train_request_params_class(
            training_data=training_data,
            poison_pills=["Bob Marley", "Bunny Livingston"],
        ),
    ).to_proto()

    caikit.core_request = build_caikit_library_request_dict(
        train_request.parameters,
        sample_lib.modules.sample_task.ListModule.TRAIN_SIGNATURE,
    )

    # model_name is not expected to be passed through
    expected_arguments = {"training_data", "poison_pills", "batch_size"}

    assert expected_arguments == set(caikit.core_request.keys())
    assert len(caikit.core_request.keys()) == 3
    assert "training_data" in caikit.core_request


def test_build_caikit_library_request_dict_works_when_data_stream_directory_includes_mixed_types(
    sample_train_service,
):
    """Global train build_caikit_library_request_dict works when mixed supported types are present"""

    with tempfile.TemporaryDirectory() as tempdir:
        fname1 = os.path.join(tempdir, "justacsv.csv")
        with open(fname1, "w") as handle:
            handle.writelines("valid_csv,1,2,3")
        fname2 = os.path.join(tempdir, "justajson.json")
        with open(fname2, "w") as handle:
            handle.writelines('{"number": 1}')

        stream_type = (
            caikit.interfaces.common.data_model.DataStreamSourceSampleTrainingType
        )
        training_data = stream_type(
            directory=stream_type.Directory(dirname=tempdir, extension="json")
        )
        train_class = DataBase.get_class_for_name("SampleTaskSampleModuleTrainRequest")
        train_request_params_class = DataBase.get_class_for_name(
            "SampleTaskSampleModuleTrainParameters"
        )
        train_request = train_class(
            model_name=random_test_id(),
            parameters=train_request_params_class(
                training_data=training_data,
            ),
        ).to_proto()

        # no error because at least 1 json file exists within the provided dir
        caikit.core_request = build_caikit_library_request_dict(
            train_request.parameters,
            sample_lib.modules.sample_task.SampleModule.TRAIN_SIGNATURE,
        )

        expected_arguments = {
            "training_data",
            "union_list",
            "batch_size",
            "oom_exit",
            "sleep_time",
            "sleep_increment",
        }

        assert expected_arguments == set(caikit.core_request.keys())
        assert "training_data" in caikit.core_request



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class with common functionality across all caikit servers"""

# Standard
from typing import Optional
import abc
import signal

# Third Party
from prometheus_client import start_http_server

# First Party
import aconfig
import alog

# Local
from caikit.config import get_config
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.service_factory import ServicePackage, ServicePackageFactory
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
import caikit

log = alog.use_channel("SERVR-BASE")


class RuntimeServerBase(abc.ABC):
    __doc__ = __doc__
    _metrics_server_started = False

    def __init__(self, base_port: int, tls_config_override: Optional[aconfig.Config]):
        self.config = get_config()
        self.port = base_port
        self.tls_config = (
            tls_config_override if tls_config_override else self.config.runtime.tls
        )
        log.debug4("Full caikit config: %s", self.config)

        # Configure using the log level and formatter type specified in config.
        caikit.core.toolkit.logging.configure()

        # We should always be able to stand up an inference service
        self.enable_inference = self.config.runtime.service_generation.enable_inference
        self.enable_training = self.config.runtime.service_generation.enable_training
        self.inference_service: Optional[ServicePackage] = (
            ServicePackageFactory.get_service_package(
                ServicePackageFactory.ServiceType.INFERENCE,
            )
            if self.enable_inference
            else None
        )

        # But maybe not always a training service
        try:
            training_service: Optional[ServicePackage] = (
                ServicePackageFactory.get_service_package(
                    ServicePackageFactory.ServiceType.TRAINING,
                )
                if self.enable_training
                else None
            )
        except CaikitRuntimeException as e:
            log.warning("Cannot stand up training service, disabling training: %s", e)
            training_service = None

        self.training_service = training_service

    @classmethod
    def _start_metrics_server(cls) -> None:
        """Start a single instance of the metrics server based on configuration"""
        if not cls._metrics_server_started and get_config().runtime.metrics.enabled:
            log.info(
                "Serving prometheus metrics on port %s",
                get_config().runtime.metrics.port,
            )
            with alog.ContextTimer(log.info, "Booted metrics server in "):
                start_http_server(get_config().runtime.metrics.port)
            cls._metrics_server_started = True

    def _intercept_interrupt_signal(self) -> None:
        """intercept signal handler"""
        signal.signal(signal.SIGINT, self.interrupt)
        signal.signal(signal.SIGTERM, self.interrupt)

    def interrupt(self, signal_, _stack_frame):
        log.info(
            "<RUN87630120I>",
            "Caikit Runtime received interrupt signal %s, shutting down",
            signal_,
        )
        self.stop()

    def _shut_down_model_manager(self):
        """Shared utility for shutting down the model manager"""
        ModelManager.get_instance().shut_down()

    @abc.abstractmethod
    def start(self, blocking: bool = True):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    # Context manager impl
    def __enter__(self):
        self.start(blocking=False)
        return self

    def __exit__(self, type_, value, traceback):
        self.stop()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# Standard
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))


# -- Project information -----------------------------------------------------

project = "Caikit"
copyright = "2023, The Caikit Authors"
author = "The Caikit Authors"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # Generate complete API docs by parsing source code
    "autoapi.extension",
    # Add links to source from generated docs
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "caikit/config"]

# -- autoapi configuration ---------------------------------------------------

# Language of source code to parse
autoapi_type = "python"

# Source code to parse to generate API docs relative to 'docs/source' directory
autoapi_dirs = [os.path.join("..", "..", "caikit")]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Support external links to specific versions of the files in the Github repo
branch = os.environ.get("READTHEDOCS_VERSION")
if branch is None or branch == "latest":
    branch = "main"

REPO = "caikit/caikit"
scm_raw_web = "https://raw.githubusercontent.com/" + REPO + branch
scm_web = "https://github.com/" + REPO + "blob/" + branch

# Store variables in the epilogue so they are globally available.
rst_epilog = """
.. |SCM_WEB| replace:: {s}
.. |SCM_RAW_WEB| replace:: {sr}
.. |SCM_BRANCH| replace:: {b}
""".format(
    s=scm_web, sr=scm_raw_web, b=branch
)

# used to have links to repo files
extlinks = {
    "scm_raw_web": (scm_raw_web + "/%s", "scm_raw_web"),
    "scm_web": (scm_web + "/%s", "scm_web"),
}



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from threading import Event, Thread
from unittest.mock import MagicMock, patch

try:
    # Standard
    from test.support.threading_helper import catch_threading_exception
except (NameError, ModuleNotFoundError):
    from tests.base import catch_threading_exception

# Standard
import time
import unittest

# Local
from caikit import get_config
from caikit.runtime.protobufs import model_runtime_pb2
from caikit.runtime.servicers.model_runtime_servicer import ModelRuntimeServicerImpl
from caikit.runtime.types.aborted_exception import AbortedException
from tests.fixtures import Fixtures


class TestModelRuntimeServicerImpl(unittest.TestCase):
    """This test suite tests the ModelRuntimeServicerImpl class"""

    def setUp(self):
        """This method runs before each test begins to run"""
        self.servicer = ModelRuntimeServicerImpl()

    def test_model_load_sets_per_model_concurrency(self):
        model = "test-any-model-id"
        # Grab a model type that has some max concurrency set
        model_type = list(
            get_config().inference_plugin.model_mesh.max_model_concurrency_per_type.keys()
        )[0]
        request = model_runtime_pb2.LoadModelRequest(
            modelId=model, modelType=model_type
        )
        context = Fixtures.build_context(model)

        expected_concurrency = (
            get_config().inference_plugin.model_mesh.max_model_concurrency_per_type[
                model_type
            ]
        )
        mock_manager = MagicMock()
        mock_manager.load_model.return_value = 1

        with patch.object(self.servicer, "model_manager", mock_manager):
            response = self.servicer.loadModel(request, context)
        self.assertEqual(expected_concurrency, response.maxConcurrency)

    def test_model_load_sets_default_max_model_concurrency(self):
        model = "test-any-model-id"
        model_type = "some-fake-model-type"
        request = model_runtime_pb2.LoadModelRequest(
            modelId=model, modelType=model_type
        )
        context = Fixtures.build_context(model)

        expected_concurrency = (
            get_config().inference_plugin.model_mesh.max_model_concurrency
        )
        mock_manager = MagicMock()
        mock_manager.load_model.return_value = 1

        with patch.object(self.servicer, "model_manager", mock_manager):
            response = self.servicer.loadModel(request, context)
        self.assertEqual(expected_concurrency, response.maxConcurrency)

    def test_load_model_aborts(self):
        """ModelRuntimeServicer.loadModel will abort a long-running load"""
        model = "test-any-model-id"
        request = model_runtime_pb2.LoadModelRequest(modelId=model)
        context = Fixtures.build_context(model)

        mock_manager = MagicMock()
        started = Event()

        def never_return(*args, **kwargs):
            started.set()
            while True:
                time.sleep(0.01)

        mock_manager.load_model.side_effect = never_return
        load_thread = Thread(target=self.servicer.loadModel, args=(request, context))

        with catch_threading_exception() as cm:
            with patch.object(self.servicer, "model_manager", mock_manager):
                load_thread.start()
                started.wait()
                context.cancel()
                load_thread.join(10)

            self.assertFalse(load_thread.is_alive())

            # Make sure the correct exception was raised
            assert cm.exc_type == AbortedException



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common interface for destroyable threads and processes
"""

# Standard
from typing import Any, Optional
import abc


class Destroyable(abc.ABC):
    __doc__ = __doc__

    @property
    @abc.abstractmethod
    def destroyed(self) -> bool:
        """Return True if destroy was called, regardless of whether the
        destroyable was alive at the time
        """

    @property
    @abc.abstractmethod
    def canceled(self) -> bool:
        """Returns True if destroyed while actively working"""

    @property
    @abc.abstractmethod
    def ran(self) -> bool:
        """Return True if the destroyable completed execution in any state"""

    @property
    @abc.abstractmethod
    def threw(self) -> bool:
        """Return True if any exception was raised during execution"""

    @abc.abstractmethod
    def get_or_throw(self) -> Any:
        """Get the result of the execution or raise an error if one occurred"""

    @abc.abstractmethod
    def destroy(self):
        """Cancel any in-progress work"""

    @property
    @abc.abstractmethod
    def error(self) -> Optional[Exception]:
        """Return the error information to user if one occurred"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model-mesh.proto
"""Generated protocol buffer code."""
# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10model-mesh.proto\x12\x05mmesh"\x81\x01\n\x14RegisterModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12#\n\tmodelInfo\x18\x02 \x01(\x0b\x32\x10.mmesh.ModelInfo\x12\x0f\n\x07loadNow\x18\x03 \x01(\x08\x12\x0c\n\x04sync\x18\x04 \x01(\x08\x12\x14\n\x0clastUsedTime\x18\x05 \x01(\x04"4\n\tModelInfo\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t"\xe4\x02\n\x0fModelStatusInfo\x12\x32\n\x06status\x18\x01 \x01(\x0e\x32".mmesh.ModelStatusInfo.ModelStatus\x12\x0e\n\x06\x65rrors\x18\x02 \x03(\t\x12<\n\x0emodelCopyInfos\x18\x03 \x03(\x0b\x32$.mmesh.ModelStatusInfo.ModelCopyInfo\x1ag\n\rModelCopyInfo\x12\x10\n\x08location\x18\x01 \x01(\t\x12\x36\n\ncopyStatus\x18\x02 \x01(\x0e\x32".mmesh.ModelStatusInfo.ModelStatus\x12\x0c\n\x04time\x18\x03 \x01(\x04"f\n\x0bModelStatus\x12\r\n\tNOT_FOUND\x10\x00\x12\x0e\n\nNOT_LOADED\x10\x01\x12\x0b\n\x07LOADING\x10\x02\x12\n\n\x06LOADED\x10\x03\x12\x12\n\x0eLOADING_FAILED\x10\x04\x12\x0b\n\x07UNKNOWN\x10\x05")\n\x16UnregisterModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"\x19\n\x17UnregisterModelResponse"#\n\x10GetStatusRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"P\n\x13\x45nsureLoadedRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12\x14\n\x0clastUsedTime\x18\x02 \x01(\x04\x12\x0c\n\x04sync\x18\x04 \x01(\x08J\x04\x08\x03\x10\x04"\xce\x02\n\x10VModelStatusInfo\x12\x34\n\x06status\x18\x01 \x01(\x0e\x32$.mmesh.VModelStatusInfo.VModelStatus\x12\x15\n\ractiveModelId\x18\x02 \x01(\t\x12\x15\n\rtargetModelId\x18\x03 \x01(\t\x12\x31\n\x11\x61\x63tiveModelStatus\x18\x04 \x01(\x0b\x32\x16.mmesh.ModelStatusInfo\x12\x31\n\x11targetModelStatus\x18\x05 \x01(\x0b\x32\x16.mmesh.ModelStatusInfo\x12\r\n\x05owner\x18\x06 \x01(\t"a\n\x0cVModelStatus\x12\r\n\tNOT_FOUND\x10\x00\x12\x0b\n\x07\x44\x45\x46INED\x10\x01\x12\x11\n\rTRANSITIONING\x10\x02\x12\x15\n\x11TRANSITION_FAILED\x10\x03\x12\x0b\n\x07UNKNOWN\x10\x05"6\n\x13\x44\x65leteVModelRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t"\x16\n\x14\x44\x65leteVModelResponse"\xef\x01\n\x10SetVModelRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\n \x01(\t\x12\x15\n\rtargetModelId\x18\x02 \x01(\t\x12\x12\n\nupdateOnly\x18\x03 \x01(\x08\x12#\n\tmodelInfo\x18\x04 \x01(\x0b\x32\x10.mmesh.ModelInfo\x12\x1d\n\x15\x61utoDeleteTargetModel\x18\x05 \x01(\x08\x12\x0f\n\x07loadNow\x18\x06 \x01(\x08\x12\r\n\x05\x66orce\x18\x07 \x01(\x08\x12\x0c\n\x04sync\x18\x08 \x01(\x08\x12\x1d\n\x15\x65xpectedTargetModelId\x18\t \x01(\t"9\n\x16GetVModelStatusRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t2\x8b\x04\n\tModelMesh\x12\x46\n\rregisterModel\x12\x1b.mmesh.RegisterModelRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12R\n\x0funregisterModel\x12\x1d.mmesh.UnregisterModelRequest\x1a\x1e.mmesh.UnregisterModelResponse"\x00\x12\x43\n\x0egetModelStatus\x12\x17.mmesh.GetStatusRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12\x44\n\x0c\x65nsureLoaded\x12\x1a.mmesh.EnsureLoadedRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12?\n\tsetVModel\x12\x17.mmesh.SetVModelRequest\x1a\x17.mmesh.VModelStatusInfo"\x00\x12I\n\x0c\x64\x65leteVModel\x12\x1a.mmesh.DeleteVModelRequest\x1a\x1b.mmesh.DeleteVModelResponse"\x00\x12K\n\x0fgetVModelStatus\x12\x1d.mmesh.GetVModelStatusRequest\x1a\x17.mmesh.VModelStatusInfo"\x00\x42(\n\x1c\x63om.ibm.watson.modelmesh.apiP\x01Z\x06/mmeshb\x06proto3'
)


_REGISTERMODELREQUEST = DESCRIPTOR.message_types_by_name["RegisterModelRequest"]
_MODELINFO = DESCRIPTOR.message_types_by_name["ModelInfo"]
_MODELSTATUSINFO = DESCRIPTOR.message_types_by_name["ModelStatusInfo"]
_MODELSTATUSINFO_MODELCOPYINFO = _MODELSTATUSINFO.nested_types_by_name["ModelCopyInfo"]
_UNREGISTERMODELREQUEST = DESCRIPTOR.message_types_by_name["UnregisterModelRequest"]
_UNREGISTERMODELRESPONSE = DESCRIPTOR.message_types_by_name["UnregisterModelResponse"]
_GETSTATUSREQUEST = DESCRIPTOR.message_types_by_name["GetStatusRequest"]
_ENSURELOADEDREQUEST = DESCRIPTOR.message_types_by_name["EnsureLoadedRequest"]
_VMODELSTATUSINFO = DESCRIPTOR.message_types_by_name["VModelStatusInfo"]
_DELETEVMODELREQUEST = DESCRIPTOR.message_types_by_name["DeleteVModelRequest"]
_DELETEVMODELRESPONSE = DESCRIPTOR.message_types_by_name["DeleteVModelResponse"]
_SETVMODELREQUEST = DESCRIPTOR.message_types_by_name["SetVModelRequest"]
_GETVMODELSTATUSREQUEST = DESCRIPTOR.message_types_by_name["GetVModelStatusRequest"]
_MODELSTATUSINFO_MODELSTATUS = _MODELSTATUSINFO.enum_types_by_name["ModelStatus"]
_VMODELSTATUSINFO_VMODELSTATUS = _VMODELSTATUSINFO.enum_types_by_name["VModelStatus"]
RegisterModelRequest = _reflection.GeneratedProtocolMessageType(
    "RegisterModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REGISTERMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.RegisterModelRequest)
    },
)
_sym_db.RegisterMessage(RegisterModelRequest)

ModelInfo = _reflection.GeneratedProtocolMessageType(
    "ModelInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelInfo)
    },
)
_sym_db.RegisterMessage(ModelInfo)

ModelStatusInfo = _reflection.GeneratedProtocolMessageType(
    "ModelStatusInfo",
    (_message.Message,),
    {
        "ModelCopyInfo": _reflection.GeneratedProtocolMessageType(
            "ModelCopyInfo",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELSTATUSINFO_MODELCOPYINFO,
                "__module__": "model_mesh_pb2"
                # @@protoc_insertion_point(class_scope:mmesh.ModelStatusInfo.ModelCopyInfo)
            },
        ),
        "DESCRIPTOR": _MODELSTATUSINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelStatusInfo)
    },
)
_sym_db.RegisterMessage(ModelStatusInfo)
_sym_db.RegisterMessage(ModelStatusInfo.ModelCopyInfo)

UnregisterModelRequest = _reflection.GeneratedProtocolMessageType(
    "UnregisterModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNREGISTERMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnregisterModelRequest)
    },
)
_sym_db.RegisterMessage(UnregisterModelRequest)

UnregisterModelResponse = _reflection.GeneratedProtocolMessageType(
    "UnregisterModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNREGISTERMODELRESPONSE,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnregisterModelResponse)
    },
)
_sym_db.RegisterMessage(UnregisterModelResponse)

GetStatusRequest = _reflection.GeneratedProtocolMessageType(
    "GetStatusRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSTATUSREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.GetStatusRequest)
    },
)
_sym_db.RegisterMessage(GetStatusRequest)

EnsureLoadedRequest = _reflection.GeneratedProtocolMessageType(
    "EnsureLoadedRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENSURELOADEDREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.EnsureLoadedRequest)
    },
)
_sym_db.RegisterMessage(EnsureLoadedRequest)

VModelStatusInfo = _reflection.GeneratedProtocolMessageType(
    "VModelStatusInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _VMODELSTATUSINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.VModelStatusInfo)
    },
)
_sym_db.RegisterMessage(VModelStatusInfo)

DeleteVModelRequest = _reflection.GeneratedProtocolMessageType(
    "DeleteVModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEVMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.DeleteVModelRequest)
    },
)
_sym_db.RegisterMessage(DeleteVModelRequest)

DeleteVModelResponse = _reflection.GeneratedProtocolMessageType(
    "DeleteVModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEVMODELRESPONSE,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.DeleteVModelResponse)
    },
)
_sym_db.RegisterMessage(DeleteVModelResponse)

SetVModelRequest = _reflection.GeneratedProtocolMessageType(
    "SetVModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SETVMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.SetVModelRequest)
    },
)
_sym_db.RegisterMessage(SetVModelRequest)

GetVModelStatusRequest = _reflection.GeneratedProtocolMessageType(
    "GetVModelStatusRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETVMODELSTATUSREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.GetVModelStatusRequest)
    },
)
_sym_db.RegisterMessage(GetVModelStatusRequest)

_MODELMESH = DESCRIPTOR.services_by_name["ModelMesh"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"\n\034com.ibm.watson.modelmesh.apiP\001Z\006/mmesh"
    )
    _REGISTERMODELREQUEST._serialized_start = 28
    _REGISTERMODELREQUEST._serialized_end = 157
    _MODELINFO._serialized_start = 159
    _MODELINFO._serialized_end = 211
    _MODELSTATUSINFO._serialized_start = 214
    _MODELSTATUSINFO._serialized_end = 570
    _MODELSTATUSINFO_MODELCOPYINFO._serialized_start = 363
    _MODELSTATUSINFO_MODELCOPYINFO._serialized_end = 466
    _MODELSTATUSINFO_MODELSTATUS._serialized_start = 468
    _MODELSTATUSINFO_MODELSTATUS._serialized_end = 570
    _UNREGISTERMODELREQUEST._serialized_start = 572
    _UNREGISTERMODELREQUEST._serialized_end = 613
    _UNREGISTERMODELRESPONSE._serialized_start = 615
    _UNREGISTERMODELRESPONSE._serialized_end = 640
    _GETSTATUSREQUEST._serialized_start = 642
    _GETSTATUSREQUEST._serialized_end = 677
    _ENSURELOADEDREQUEST._serialized_start = 679
    _ENSURELOADEDREQUEST._serialized_end = 759
    _VMODELSTATUSINFO._serialized_start = 762
    _VMODELSTATUSINFO._serialized_end = 1096
    _VMODELSTATUSINFO_VMODELSTATUS._serialized_start = 999
    _VMODELSTATUSINFO_VMODELSTATUS._serialized_end = 1096
    _DELETEVMODELREQUEST._serialized_start = 1098
    _DELETEVMODELREQUEST._serialized_end = 1152
    _DELETEVMODELRESPONSE._serialized_start = 1154
    _DELETEVMODELRESPONSE._serialized_end = 1176
    _SETVMODELREQUEST._serialized_start = 1179
    _SETVMODELREQUEST._serialized_end = 1418
    _GETVMODELSTATUSREQUEST._serialized_start = 1420
    _GETVMODELSTATUSREQUEST._serialized_end = 1477
    _MODELMESH._serialized_start = 1480
    _MODELMESH._serialized_end = 2003
# @@protoc_insertion_point(module_scope)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from collections import Counter as DictCounter
from functools import partial
from typing import Dict, Optional
import atexit
import gc
import os
import threading

# Third Party
from grpc import StatusCode
from prometheus_client import Counter, Gauge, Summary

# First Party
import alog

# Local
from caikit import get_config
from caikit.core import ModuleBase
from caikit.core.exceptions import error_handler
from caikit.runtime.model_management.loaded_model import LoadedModel
from caikit.runtime.model_management.model_loader import ModelLoader
from caikit.runtime.model_management.model_sizer import ModelSizer
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.work_management.abortable_action import ActionAborter

log = alog.use_channel("MODEL-MANAGR")
error = error_handler.get(log)

MODEL_SIZE_GAUGE = Gauge(
    "total_loaded_models_size",
    "Total size of loaded models reported to model-mesh",
    ["model_type"],
)
MODEL_COUNT_GAUGE = Gauge(
    "total_loaded_models", "Total number of loaded models", ["model_type", "model_id"]
)
LOAD_MODEL_EXCEPTION_COUNTER = Counter(
    "load_model_exception_count",
    "Count of exceptions raised during loadModel RPCs",
    ["model_type"],
)
LOAD_MODEL_DURATION_SUMMARY = Summary(
    "load_model_duration_seconds",
    "Summary of the duration (in seconds) of loadModel RPCs",
    ["model_type"],
)
LOCAL_MODEL_TYPE = "LOCAL"


class ModelManager:  # pylint: disable=too-many-instance-attributes
    """Model Manager class. The singleton class contains the core implementational details
    for the Model Runtime (i.e., load/unload functionality, etc). It does not provide the core
    details for predict calls."""

    __instance = None

    __model_size_gauge_lock = threading.Lock()

    _LOCAL_MODEL_TYPE = "standalone-model"

    ## Construction ##

    @classmethod
    def get_instance(cls) -> "ModelManager":
        """This method returns the instance of Model Manager"""
        if not cls.__instance:
            cls.__instance = ModelManager()
        return cls.__instance

    def __init__(self):
        """Initialize a ModelManager instance."""
        # Re-instantiating this is a programming error
        assert self.__class__.__instance is None, "This class is a singleton!"
        ModelManager.__instance = self

        # Pull in a ModelLoader and ModelSizer
        self.model_loader = ModelLoader.get_instance()
        self.model_sizer = ModelSizer.get_instance()

        # In-memory mapping of model_id to LoadedModel instance
        self.loaded_models: Dict[str, LoadedModel] = {}

        # Lock for mutating operations on loaded_models
        self._loaded_models_lock = threading.Lock()

        # Optionally load models mounted into a local directory
        runtime_cfg = get_config().runtime
        self._local_models_dir = runtime_cfg.local_models_dir or ""
        if self._local_models_dir and not os.path.exists(self._local_models_dir):
            log.warning(
                "<RUN53709826W>",
                "Invalid runtime.local_models_dir %s. Does not exist",
                self._local_models_dir,
            )
            self._local_models_dir = ""

        # Keep track of whether lazy loading is enabled
        self._lazy_load_local_models = runtime_cfg.lazy_load_local_models
        error.value_check(
            "<RUN44773514E>",
            not self._lazy_load_local_models or self._local_models_dir,
            "Must set runtime.local_models_dir with runtime.lazy_load_local_models",
        )

        # Set up local model periodic sync
        self._lazy_load_poll_period_seconds = runtime_cfg.lazy_load_poll_period_seconds
        error.type_check(
            "<RUN59138047E>",
            int,
            float,
            allow_none=True,
            lazy_load_poll_period_seconds=self._lazy_load_poll_period_seconds,
        )
        self._lazy_sync_timer = None
        self._enable_lazy_load_poll = (
            self._local_models_dir
            and self._lazy_load_local_models
            and self._lazy_load_poll_period_seconds
        )
        if self._enable_lazy_load_poll:
            atexit.register(self.shut_down)

        # Do the initial local models load
        if self._local_models_dir:
            log.info("<RUN44739400I>", "Loading local models into Caikit Runtime...")
            self.sync_local_models(wait=True)

    def shut_down(self):
        """Shut down cache purging"""
        self._enable_lazy_load_poll = False
        timer = getattr(self, "_lazy_sync_timer", None)
        if timer is not None:
            timer.cancel()
            if timer.is_alive():
                timer.join()

    ## Model Management ##

    def load_model(
        self,
        model_id: str,
        local_model_path: str,
        model_type: str,
        wait: bool = True,
        aborter: Optional[ActionAborter] = None,
        retries: Optional[int] = None,
    ) -> int:
        """Load a model using model_path (in Cloud Object Storage) & give it a model ID
        Args:
            model_id (str):  Model ID string for the model to load.
            local_model_path (str): Local path to load the model from.
            model_type (str): Type of the model to load.
            wait (bool): Wait for the model to finish loading
            aborter (Optional[ActionAborter]): The aborter to use for this load
            retries: Optional[int]: Number of times to retry on load failure
        Returns:
            Model_size (int) : Size of the loaded model in bytes
        """
        with LOAD_MODEL_DURATION_SUMMARY.labels(model_type=model_type).time():

            # If already loaded, just return the size
            # NOTE: We make the dict access atomic here to avoid the race where
            #   we check if model_id in the map, then re-look it up to get the
            #   size which could fail if it is unloaded between the two.
            model = self.loaded_models.get(model_id)
            if model is not None:
                log.debug("Model '%s' is already loaded", model_id)
                return model.size()

            # Grab the mutation lock and load the model if needed
            with self._loaded_models_lock:
                # Re-check now that the mutation lock is held
                model = self.loaded_models.get(model_id)
                if model is None:
                    log.debug("Loading %s from %s", model_id, local_model_path)
                    try:
                        model = self.model_loader.load_model(
                            model_id,
                            local_model_path,
                            model_type,
                            aborter=aborter,
                            fail_callback=partial(self.unload_model, model_id),
                            retries=retries,
                        )
                    except Exception as ex:
                        self.__increment_load_model_exception_count_metric(model_type)
                        raise ex

                    # Estimate the model's size and update the LoadedModel
                    try:
                        model_size = self.model_sizer.get_model_size(
                            model_id, local_model_path, model_type
                        )
                    except CaikitRuntimeException:
                        log.debug(
                            "Unable to estimate model size for non-disk model: %s",
                            model_id,
                        )
                        model_size = 0
                    model.set_size(model_size)

                    # Add model + helpful metadata to our loaded models map
                    self.loaded_models[model_id] = model

                    # Update Prometheus metrics
                    self.__increment_model_count_metric(model_type, model_id)
                    self.__report_total_model_size_metric()

            # If waiting, do so outside of the mutation lock
            if wait:
                model.wait()

            # Return the model's size
            return model.size()

    def sync_local_models(self, wait: bool = False):
        """Sync in-memory models with models in the configured local_model_dir

        New models will be loaded and models previously loaded from local will
        be unloaded.

        Args:
            wait (bool): Wait for loading to complete
        """
        try:
            self._local_models_dir_sync(wait)
        except StopIteration:
            log.warning(
                "<RUN56519883W>",
                "local_models_dir %s unreachable. Terminating synchronization",
                self._local_models_dir,
            )
            self._enable_lazy_load_poll = False
        except Exception as err:  # pylint: disable=broad-exception-caught
            log.warning(
                "<RUN44524933W>",
                "Exception raised during local_models_dir sync: %s",
                str(err),
                exc_info=True,
            )

        # If running periodically, kick off the next iteration
        if self._enable_lazy_load_poll:
            if self._lazy_sync_timer is None:
                log.info(
                    "Initializing local_models_dir sync with period %s",
                    self._lazy_load_poll_period_seconds,
                )
            if self._lazy_sync_timer is not None and self._lazy_sync_timer.is_alive():
                log.debug3("Canceling live timer")
                self._lazy_sync_timer.cancel()
            log.debug3(
                "Starting next poll timer for %ss", self._lazy_load_poll_period_seconds
            )
            log.debug4(
                "All open threads: %s",
                [thread.name for thread in threading.enumerate()],
            )
            self._lazy_sync_timer = threading.Timer(
                self._lazy_load_poll_period_seconds, self.sync_local_models
            )
            self._lazy_sync_timer.daemon = True
            self._lazy_sync_timer.start()

    def unload_model(self, model_id) -> int:
        """Unload a model by ID model.

        Args:
            model_id (string):  Model ID string for the model to unload. If None,
                (default) the default model id will be used.
        Returns:
            Model_size (int) : Size of the loaded model in bytes
        """
        log.debug("List of loaded models: %s", str(self.loaded_models))
        # If the model failed to load, just return 0; no need to throw an error here.
        if model_id not in self.loaded_models:
            log.debug("Model '%s' is not loaded, so it cannot be unloaded!", model_id)
            return 0

        # Temporarily store model size and type info
        model_type = self.loaded_models[model_id].type()
        model_size = self.loaded_models[model_id].size()

        # Delete the model and remove it from the model map
        try:
            model = self.loaded_models.pop(model_id)
            # If the model is still loading, we need to wait for it to finish so
            # that we can do our best to fully free it
            model.wait()
            del model
        except CaikitRuntimeException:
            raise
        except Exception as ex:
            log.debug("Model '%s' failed to unload with error: %s", model_id, repr(ex))
            raise CaikitRuntimeException(
                StatusCode.INTERNAL,
                "Model could not be unloaded!!",
                {"model_id": model_id},
            ) from ex

        # Invoke the garbage collector to ensure that memory is freed and ready for loading new
        # models.  This also helps troubleshooting when examining memory usage during unload.
        gc.collect()

        # Update Prometheus metrics
        self.__report_total_model_size_metric()
        self.__decrement_model_count_metric(model_type, model_id)

        return model_size

    def unload_all_models(self):
        """Unload all loaded models"""
        all_model_ids = list(self.loaded_models.keys())
        for model_id in all_model_ids:
            self.unload_model(model_id)

    def get_model_size(self, model_id) -> int:
        """Look up size of a model by model ID.
        Args:
            model_id (string):  Model ID string for the model. Throw Exception if empty,
                or it is an ID of a model that is not currently loaded.
        Returns:
            Model_size (int) : Size of the loaded model in bytes
        """
        if not model_id or model_id not in self.loaded_models:
            msg = (
                "Unable to retrieve the size of model '%s'; it is unregistered or unloaded."
                % model_id
            )
            log.debug(msg)
            raise CaikitRuntimeException(
                StatusCode.NOT_FOUND, msg, {"model_id": model_id}
            )

        loaded_model = self.loaded_models[model_id]
        self.__report_total_model_size_metric()
        return loaded_model.size()

    def estimate_model_size(self, model_id, local_model_path, model_type) -> int:
        """Predict size of a model using model ID and path.
        Args:
            model_id (string):  Model ID string for the model to predict size of.
            local_model_path (string): Local path to the model.
            model_type (string): Type of the model
        Returns:
            Model_size (int) : Estimated size of the model in bytes.
        """
        return self.model_sizer.get_model_size(model_id, local_model_path, model_type)

    def retrieve_model(self, model_id: str) -> ModuleBase:
        """Retrieve a model from the loaded model map by model ID.

        Args:
            model_id (str): Model ID of the model to retrieve
        Returns:
            response (caikit.core.module.ModuleBase):
                A loaded Caikit model
        """
        if not model_id or not isinstance(model_id, str):
            raise CaikitRuntimeException(
                StatusCode.INVALID_ARGUMENT, "Missing required model ID"
            )

        # Now retrieve the model and fall back to lazy loading
        model_loaded = model_id in self.loaded_models
        if not model_loaded and self._lazy_load_local_models:
            local_model_path = os.path.join(self._local_models_dir, model_id)
            log.debug2(
                "Lazy loading local model %s from %s", model_id, local_model_path
            )
            # If the model is not present on disk, attempt to lazy load it
            # anyway using the model_id as the "model_path". This allows
            # auto-finders that can infer the model's config to load based on
            # the ID.
            if not os.path.exists(local_model_path):
                log.debug2("Attempting to load ephemeral model %s", model_id)
                local_model_path = model_id
            self.load_model(
                model_id=model_id,
                local_model_path=local_model_path,
                model_type=self._LOCAL_MODEL_TYPE,
                wait=True,
                retries=get_config().runtime.lazy_load_retries,
            )
            model_loaded = True

        # If still not loaded, there's nothing to find, so raise NOT_FOUND
        if not model_loaded:
            msg = f"Model '{model_id}' not loaded"
            log.debug(
                {"log_code": "<RUN61105243D>", "message": msg, "model_id": model_id}
            )
            raise CaikitRuntimeException(
                StatusCode.NOT_FOUND, msg, {"model_id": model_id}
            )

        # NOTE: If the model is partially loaded, this call will wait on the
        #   model future in the LoadedModel
        return self.loaded_models[model_id].model()

    ## Implementation Details ##

    def _local_models_dir_sync(self, wait: bool = False):
        """This function implements the mechanics of synchronizing the
        local_models_dir and the in-memory loaded_models map. It may raise and
        therefore errors should be handled by the wrapper function.

        NOTE: In the case that the local_models_dir becomes unreadable, it will
            raise StopIteration to indicate that any periodic synchronization
            should terminate.
        """
        # Get the list of models on disk
        # NOTE: If the local_models_dir has disappeared, this is likely a unit
        #   test with a temp dir, but in any event, we should stop trying to
        #   sync going forward
        try:
            disk_models = os.listdir(self._local_models_dir)
        except FileNotFoundError as err:
            raise StopIteration() from err

        # Find all models that aren't currently loaded
        new_models = [
            model_id for model_id in disk_models if model_id not in self.loaded_models
        ]
        log.debug("New local models: %s", new_models)

        # Find all models that are currently loaded from the local models dir
        # that no longer exist
        unload_models = [
            model_id
            for model_id, loaded_model in self.loaded_models.items()
            if model_id not in disk_models
            and loaded_model.path().startswith(
                self._local_models_dir,
            )
        ]
        log.debug("Unloaded local models: %s", unload_models)

        # Load new models
        for model_id in new_models:
            model_path = os.path.join(self._local_models_dir, model_id)
            self.load_model(
                model_id,
                model_path,
                self._LOCAL_MODEL_TYPE,
                wait=False,
                retries=get_config().runtime.lazy_load_retries,
            )

        # Unload old models
        # NOTE: No need for error handling here since unload_model will warn on
        #   errors and move on
        for model_id in unload_models:
            log.debug2("Unloading local model %s", model_id)
            self.unload_model(model_id)

        # Wait for models to load
        if wait:
            for model_id in new_models:
                loaded_model = self.loaded_models.get(model_id)
                # If somehow already purged, there's nothing to wait on. This is
                # extremely unlikely since it would require another thread to
                # explicitly call unload on the model AND have the model finish
                # loading between then and now. Better to be safe than sorry,
                # though!
                if loaded_model is None:  # pragma: no cover
                    continue
                # Wait for it and make sure it didn't fail
                try:
                    loaded_model.wait()
                except CaikitRuntimeException as err:
                    log.warning(
                        "<RUN56627485W>",
                        "Failed to load model %s: %s",
                        model_id,
                        repr(err),
                        exc_info=True,
                    )

    def __report_total_model_size_metric(self):
        # Just a happy little lock to ensure that with concurrent loading and unloading,
        # the last metric reported will be correct.
        with self.__model_size_gauge_lock:
            cnt = DictCounter()
            for model in self.loaded_models.values():
                cnt[model.type()] += model.size()

            for model_type, total_size in cnt.items():
                MODEL_SIZE_GAUGE.labels(model_type=model_type).set(total_size)

    @staticmethod
    def __increment_model_count_metric(model_type, model_id):
        MODEL_COUNT_GAUGE.labels(model_type=model_type, model_id=model_id).inc()

    @staticmethod
    def __decrement_model_count_metric(model_type, model_id):
        MODEL_COUNT_GAUGE.labels(model_type=model_type, model_id=model_id).dec()

    @staticmethod
    def __increment_load_model_exception_count_metric(model_type):
        LOAD_MODEL_EXCEPTION_COUNTER.labels(model_type=model_type).inc()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import io
import json
import os
import tempfile

# Third Party
import pytest
import yaml

# Local
# Unit Test Infrastructure
from tests.base import TestCaseBase
import caikit.core


class TestCoreSerializers(TestCaseBase):
    def test_json_serializer(self):
        with tempfile.TemporaryDirectory() as tempdir:
            input_object = {"foo": "bar"}

            serializer = caikit.core.toolkit.serializers.JSONSerializer()
            save_path = os.path.join(tempdir, "test.json")
            serializer.serialize(input_object, save_path)

            self.assertTrue(os.path.exists(save_path))
            with open(save_path, "r") as handle:
                self.assertEqual(input_object, json.load(handle))

    def test_text_serializer(self):
        with tempfile.TemporaryDirectory() as tempdir:
            input_list = ["foo", "bar"]
            serializer = caikit.core.toolkit.serializers.TextSerializer()
            save_path = os.path.join(tempdir, "test.txt")
            serializer.serialize(input_list, save_path)

            self.assertTrue(os.path.exists(save_path))
            with open(save_path, mode="r", encoding="utf8") as fh:
                output_list = list(map(str.strip, fh.readlines()))
            self.assertEqual(input_list, output_list)

    def test_yaml_serializer(self):
        with tempfile.TemporaryDirectory() as tempdir:
            input_object = {"foo": "bar"}
            serializer = caikit.core.toolkit.serializers.YAMLSerializer()
            save_path = os.path.join(tempdir, "test.yaml")
            serializer.serialize(input_object, save_path)

            self.assertTrue(os.path.exists(save_path))
            with open(save_path, "r") as handle:
                self.assertEqual(input_object, yaml.safe_load(handle))

    def test_not_implemented_error(self):
        """Make sure that a NotImplementedError is raised if the base class is
        invoked directly
        """
        with pytest.raises(TypeError):
            caikit.core.toolkit.serializers.ObjectSerializer()
